# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Trial Graph Module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import time
import weakref

from collections import namedtuple, defaultdict
from functools import partial

from future.utils import viewitems, viewvalues, viewkeys

from .structures import Single, Call, Group, Mixed, TreeElement, prepare_cache
from .structures import Graph


Edge = namedtuple("Edge", "node count")


class TreeVisitor(object):
    """Create Dict Tree from Intermediate Tree"""

    def __init__(self):
        self.nodes = []
        self.edges = []
        self.delegated = {
            "initial": Edge(0, 1)
        }
        self.nid = 0
        self.min_duration = defaultdict(partial(int, 1000 ** 10))
        self.max_duration = defaultdict(partial(int, 0))
        self.keep = None

    def update_durations(self, duration, tid):
        """Update min and max duration"""
        self.max_duration[tid] = max(self.max_duration[tid], duration)
        self.min_duration[tid] = min(self.min_duration[tid], duration)

    def update_node(self, node):                                                 # pylint: disable=no-self-use
        """Update node with info and mean duration"""
        node["mean"] = node["duration"] / node["count"]
        node["info"].update_by_node(node)
        node["info"] = repr(node["info"])

    def to_dict(self):
        """Convert graph to dict"""
        for node in self.nodes:
            nnode = node["node"]
            self.update_node(nnode)
            self.update_durations(nnode["duration"], nnode["trial_id"])

        self.update_edges()
        return {
            "nodes": self.nodes,
            "edges": self.edges,
            "min_duration": self.min_duration,
            "max_duration": self.max_duration
        }

    def update_edges(self):
        """Update edges"""
        pass

    def add_node(self, node):
        """Add node to result"""
        self.nodes.append(node.to_dict(self.nid))
        original = self.nid
        self.nid += 1
        return original

    def add_edge(self, source, target, count, typ):
        """Add edge to result"""
        self.edges.append({
            "source": source,
            "target": target,
            "count": count,
            "type": typ
        })

    def visit_call(self, call):
        """Visit Call Node (Visitor Pattern)"""
        caller_id = self.add_node(call.caller)
        self.nodes[caller_id]["repr"] = repr(call)
        callees = call.called.visit(self)
        pos = 1
        for callee_id in callees:
            self.add_edge(caller_id, callee_id, pos, "call")
            pos += 1
        return [caller_id]

    def visit_group(self, group):
        """Visit Group Node (Visitor Pattern)"""
        result = []
        for element in viewvalues(group.nodes):
            result += element.visit(self)
        return result

    def visit_single(self, single):
        """Visit Single Node (Visitor Pattern)"""
        return [self.add_node(single)]

    def visit_mixed(self, mixed):
        """Visit Mixed Node (Visitor Pattern)"""
        mixed.mix_results()
        node_id = mixed.first.visit(self)
        self.nodes[node_id[0]]["duration"] = mixed.duration
        return node_id

    def visit_treeelement(self, tree_element):                                   # pylint: disable=no-self-use, unused-argument
        """Visit TreeElement Node (Visitor Pattern)"""
        return []


class NoMatchVisitor(TreeVisitor):
    """Create Dict No Match from Intermediate Tree"""
    def update_edges(self):
        for edge in self.edges:
            if edge["type"] in ["return", "call"]:
                edge["count"] = ""

    def use_delegated(self):
        """Process delegated edge"""
        result = self.delegated
        self.delegated = {}
        return result

    def solve_delegation(self, node_id, node_count, delegated):
        """Solve edge delegation"""
        self.solve_cis_delegation(node_id, node_count, delegated)
        self.solve_ret_delegation(node_id, node_count, delegated)

    def solve_cis_delegation(self, node_id, node_count, delegated):
        """Solve edge (call, initial, sequence) delegation"""
        # call initial sequence
        for typ in ["call", "initial", "sequence"]:
            if typ in delegated:
                edge = delegated[typ]
                self.add_edge(edge.node, node_id, node_count, typ)

    def solve_ret_delegation(self, node_id, node_count, delegated):              # pylint: disable=unused-argument
        """Solve edge (return) delegation"""
        if not self.nodes[node_id]["node"]["finished"]:
            return
        if "return" in delegated:
            edge = delegated["return"]
            self.add_edge(node_id, edge.node, edge.count, "return")

    def visit_call(self, call):
        delegated = self.use_delegated()
        caller_id = self.add_node(call.caller)
        self.nodes[caller_id]["repr"] = repr(call)

        if delegated:
            self.solve_delegation(caller_id, call.count, delegated)

        self.delegated["call"] = Edge(caller_id, 1)
        self.delegated["return"] = Edge(caller_id, 1)

        call.called.visit(self)
        return caller_id, call

    def visit_group(self, group):
        delegated = self.use_delegated()

        node_map = {}
        for element in viewvalues(group.nodes):
            node_id, node = element.visit(self)
            node_map[node] = node_id

        self.solve_cis_delegation(node_map[group.next_element],
                                  group.count, delegated)
        self.solve_ret_delegation(node_map[group.last], group.count, delegated)

        for previous, edges in viewitems(group.edges):
            for nnext, count in viewitems(edges):
                self.add_edge(node_map[previous], node_map[nnext],
                              count, "sequence")

        return node_map[group.next_element], group.next_element

    def visit_single(self, single):
        delegated = self.use_delegated()
        node_id = self.add_node(single)
        self.nodes[node_id]["repr"] = repr(single)

        if delegated:
            self.solve_delegation(node_id, single.count, delegated)
        return node_id, single

    def visit_mixed(self, mixed):
        mixed.mix_results()
        node_id, node = mixed.first.visit(self)
        self.nodes[node_id]["duration"] = mixed.duration
        return node_id, node

    def visit_treeelement(self, tree_element):
        return None, tree_element


class ExactMatchVisitor(NoMatchVisitor):
    """Create Dict Exact Match from Intermediate Tree"""

    def visit_single(self, single):
        _single = Single(single.activation)
        _single.level = single.level
        _single.use_id = False
        return _single

    def visit_mixed(self, mixed):
        mixed.use_id = False
        _mixed = Mixed(mixed.elements[0].visit(self))
        for element in mixed.elements[1:]:
            _mixed.add_element(element.visit(self))
        _mixed.level = mixed.level
        return _mixed

    def visit_group(self, group):
        nodes = list(viewkeys(group.nodes))
        _group = Group()
        _group.use_id = False
        _group.initialize(nodes[1].visit(self), nodes[0].visit(self))
        for element in nodes[2:]:
            _group.add_subelement(element.visit(self))
        _group.level = group.level
        return _group

    def visit_call(self, call):
        caller = call.caller.visit(self)
        called = call.called.visit(self)
        _call = Call(caller, called)
        _call.use_id = False
        _call.level = call.level
        return _call

    def visit_treeelement(self, tree_element):
        return tree_element


def update_namespace_node(node, single):
    """Update node information"""
    node["count"] += single.count
    node["duration"] += single.duration
    node["info"].add_activation(single.activation)


class NamespaceVisitor(NoMatchVisitor):
    """Create Dict Namespace from Intermediate Tree"""

    def __init__(self):
        super(NamespaceVisitor, self).__init__()
        self.context = {}
        self.context_edges = {}
        self.namestack = []

    def update_edges(self):
        pass

    def namespace(self):
        """Return current namespace"""
        return " ".join(self.namestack)

    def add_node(self, single):
        self.namestack.append(single.name_id())
        namespace = self.namespace()
        self.namestack.pop()
        if namespace in self.context:
            context = self.context[namespace]
            update_namespace_node(context["node"], single)

            return self.context[namespace]["index"]

        single.namespace = namespace
        result = super(NamespaceVisitor, self).add_node(single)
        self.context[namespace] = self.nodes[-1]
        return result

    def add_edge(self, source, target, count, typ):

        edge = "{} {} {}".format(source, target, typ)
        if edge not in self.context_edges:
            super(NamespaceVisitor, self).add_edge(source, target,
                                                   count, typ)
            self.context_edges[edge] = self.edges[-1]
        else:
            _edge = self.context_edges[edge]
            _edge["count"] += count

    def visit_call(self, call):
        self.namestack.append(call.caller.name_id())
        result = super(NamespaceVisitor, self).visit_call(call)
        self.namestack.pop()
        return result

    def visit_mixed(self, mixed):
        node_id, node = None, None
        for element in mixed.elements:
            node_id, node = element.visit(self)
        return node_id, node


def sequence(previous, nnext):
    """Create Group or add Node to it"""
    if isinstance(nnext, Group):
        nnext.add_subelement(previous)
        return nnext
    return Group().initialize(previous, nnext)


def create_group(group_list):
    """Transform a list of Single activations into a group"""
    _next = group_list.pop()
    while group_list:
        _previous = group_list.pop()
        _next = sequence(_previous, _next)
    return _next


def recursive_generate_graph(trial, single, depth):
    """Generate Graph up to the depth"""
    if not depth:
        return single
    children = []
    for act in single.activation.children:
        child = Single(act)
        child.level = single.level + 1
        children.append(recursive_generate_graph(trial, child, depth - 1))

    if children:
        group = create_group(children)
        call = Call(single, group)
        call.level = single.level
        group.level = single.level + 1
        return call

    return single


def generate_graph(trial, depth=1000):
    """Return the activation graph"""
    activations = list(trial.initial_activations)
    if not activations:
        tree = TreeElement(level=0)
        tree.trial_id = trial.id
        return tree

    return recursive_generate_graph(trial, Single(activations[0]), depth - 1)


cache = prepare_cache(                                                           # pylint: disable=invalid-name
    lambda self, *args, **kwargs: "trial {}".format(self.trial.id))


class TrialGraph(Graph):
    """Trial Graph Class
       Present trial graph on Jupyter"""

    def __init__(self, trial):
        self._graph = None
        self.trial = weakref.proxy(trial)

        self.use_cache = True
        self.width = 500
        self.height = 500
        self.mode = 3
        self._modes = {
            0: self.tree,
            1: self.no_match,
            2: self.exact_match,
            3: self.namespace_match
        }

    @cache("graph")
    def graph(self):
        """Generate an activation tree structure"""
        if self._graph is None:
            self._graph = generate_graph(self.trial)
        return self.trial.finished, self._graph

    @cache("tree")
    def tree(self):
        """Convert tree structure into dict tree structure"""
        finished, graph = self.graph()
        visitor = TreeVisitor()
        graph.visit(visitor)
        return finished, visitor.to_dict()

    @cache("no_match")
    def no_match(self):
        """Convert tree structure into dict graph without node matchings"""
        finished, graph = self.graph()
        visitor = NoMatchVisitor()
        graph.visit(visitor)
        return finished, visitor.to_dict()

    @cache("exact_match")
    def exact_match(self):
        """Convert tree structure into dict graph and match equal calls"""
        finished, graph = self.graph()
        graph = graph.visit(ExactMatchVisitor())
        visitor = NoMatchVisitor()
        graph.visit(visitor)
        return finished, visitor.to_dict()

    @cache("namespace_match")
    def namespace_match(self):
        """Convert tree structure into dict graph and match namespaces"""
        finished, graph = self.graph()
        visitor = NamespaceVisitor()
        graph.visit(visitor)
        return finished, visitor.to_dict()

    def _repr_html_(self):
        """Display d3 graph on jupyter notebook"""
        uid = str(int(time.time() * 1000000))

        result = """
            <div class="nowip-trial" data-width="{width}"
                 data-height="{height}" data-uid="{uid}"
                 data-id="{id}">
                {data}
            </div>
        """.format(
            uid=uid, id=self.trial.id,
            data=self.escape_json(self._modes[self.mode]()[1]),
            width=self.width, height=self.height)
        return result



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Execution provenance collector"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import traceback
import weakref

from ...utils.cross_version import cross_compile
from ...utils.io import print_msg
from ...utils.metaprofiler import meta_profiler

from .debugger import debugger_builtins
from .profiler import Profiler
from .slicing import Tracer                                                      # pylint: disable=unused-import


Tracker = Tracer


class Execution(object):
    """Execution Class"""

    def __init__(self, metascript):
        self.metascript = weakref.proxy(metascript)
        self.provider = None
        self.partial = False
        self.force_msg = False
        self.msg = ""

    def set_provider(self):
        """Set execution provenance provider"""
        metascript = self.metascript
        glob = globals()
        provider_cls = glob.get(metascript.execution_provenance, Profiler)
        self.provider = provider_cls(metascript)

    # ToDo #76: Processor load. Should be collected from time to time
    #                         (there are static and dynamic metadata)
    # print os.getloadavg()
    @meta_profiler("execution")
    def collect_provenance(self):
        """Collect execution provenance"""
        metascript = self.metascript
        self.set_provider()

        if metascript.compiled is None:
            metascript.compiled = cross_compile(
                metascript.code, metascript.path, "exec"
            )
        debugger_builtins(
            self.provider, metascript.namespace["__builtins__"], metascript
        )

        print_msg("  executing the script")
        self.provider.tearup()  # It must be right before exec
        try:
            exec(metascript.compiled, metascript.namespace)                      # pylint: disable=exec-used
            if '__doc__' in metascript.namespace:
                metascript.docstring = metascript.namespace['__doc__']
        except SystemExit as ex:
            self.force_msg = self.partial = ex.code > 0
            self.msg = ("the execution exited via sys.exit(). Exit status: {}"
                        "".format(ex.code))
        except Exception as exc:                                                 # pylint: disable=broad-except
            # ToDo #77: exceptions should be registered as return from the
            # activation and stored in the database. We are currently ignoring
            # all the activation tree when exceptions are raised.
            self.force_msg = self.partial = True
            self.msg = ("{}\n"
                        "the execution finished with an uncaught exception. {}"
                        "".format(repr(exc), traceback.format_exc()))
        else:
            self.force_msg = self.partial = False
            self.msg = ("the execution of trial {} finished successfully"
                        "".format(metascript.trial_id))

    @meta_profiler("storage")
    def store_provenance(self):
        """Disable provider and store provenance"""
        self.provider.teardown()
        self.provider.store(partial=self.partial)
        if self.msg:
            print_msg(self.msg, self.force_msg)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Define Metadata classes to be used as configuration """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import fnmatch
import os
import sys

from datetime import datetime

from pyposast import native_decode_source

from ..persistence import persistence_config, get_serializer
from ..persistence.lightweight import ObjectStore
from ..persistence.lightweight import DefinitionLW, ObjectLW
from ..persistence.lightweight import EnvironmentAttrLW
from ..persistence.lightweight import ModuleLW, DependencyLW
from ..persistence.lightweight import ActivationLW, ObjectValueLW
from ..persistence.lightweight import FileAccessLW, VariableLW
from ..persistence.lightweight import VariableUsageLW, VariableDependencyLW
from ..utils import io

from .prov_definition.definition import Definition
from .prov_execution.execution import Execution
from .prov_deployment.deployment import Deployment


LAST_TRIAL = ".last_trial"
MAIN = 0
PACKAGE = 1
ALL = 2

CONTEXTS = {
    "main": MAIN,
    "package": PACKAGE,
    "all": ALL
}


class Metascript(object):                                                        # pylint: disable=too-many-instance-attributes
    """Metascript object. Contain storages and arguments"""

    def __init__(self):
        # Storage
        self.definitions_store = ObjectStore(DefinitionLW)
        self.objects_store = ObjectStore(ObjectLW)

        self.environment_attrs_store = ObjectStore(EnvironmentAttrLW)
        self.modules_store = ObjectStore(ModuleLW)
        self.dependencies_store = ObjectStore(DependencyLW)

        self.activations_store = ObjectStore(ActivationLW)
        self.object_values_store = ObjectStore(ObjectValueLW)
        self.file_accesses_store = ObjectStore(FileAccessLW)

        self.variables_store = ObjectStore(VariableLW)
        self.variables_dependencies_store = ObjectStore(VariableDependencyLW)
        self.usages_store = ObjectStore(VariableUsageLW)

        # Definition object : Definition
        self.definition = Definition(self)
        self.execution = Execution(self)
        self.deployment = Deployment(self)

        # Trial id read from Database : int
        self.trial_id = None
        # Trial name : str
        self.name = None
        # Source code of the main script : str
        self.code = None
        # Compiled code : types.CodeType
        self.compiled = None
        # Script dir : str
        self.dir = None
        # Main path : str
        self._path = None
        # Map of paths considered to their def : dict(str -> DefinitionLW)
        self.paths = {}
        # Main namespace : dict
        self.namespace = None
        # Argv
        self.argv = None
        # Object Serialize function
        self.serialize = None
        # Script docstring
        self.docstring = ""

        # Verbose print
        self.verbose = False
        # Should it create a file with the last executed trial id : bool
        self.should_create_last_file = False
        # Profile noWorkflow itself
        self.meta = False

        # Show script disassembly : bool
        self.disasm = False
        # Show script disassembly before changes : bool
        self.disasm0 = False

        # Bypass module check : bool
        self.bypass_modules = False

        # Depth for capturing function activations : int
        self.depth = sys.getrecursionlimit()
        # Depth for capturing function activations outside context : int
        self.non_user_depth = 1
        # Execution provenance provider : str
        self.execution_provenance = "Profiler"
        # Script context : ["main", "package", "all"]
        self._context = MAIN
        # Save every X ms : int
        self.save_frequency = 1000
        # Save after closing X activations
        self.call_storage_frequency = 0

        # Passed arguments : str
        self.command = ""


    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    @property
    def path(self):
        """Return path"""
        return self._path

    @property
    def code_hash(self):
        """Return code_hash of trial script"""
        return self.paths[self.path].code_hash

    @path.setter
    def path(self, path):
        """Set _path.
        Remove old path from paths list and add new one to it
        """

        if self._path:
            self.definitions_store.remove(self.paths[self._path])
            del self.paths[self._path]
        self._path = path
        if not path:
            return
        self.add_path(path, set_code=True)

    def add_path(self, path, set_code=False):
        """Add path to paths list"""
        with open(path, "rb") as script_file:
            code = native_decode_source(script_file.read())
            if set_code:
                self.code = code
            self.paths[path] = self.definitions_store.dry_add(
                "", path, code, "FILE", None, 0, 0, "")

    def fake_path(self, path, code):
        """Fake configuration for tests"""
        self.name = path
        self._path = path
        self.code = native_decode_source(code)
        self.paths[path] = self.definitions_store.dry_add(
            "", path, self.code, "FILE", None, 0, 0, "")

    @property
    def context(self):
        """Return context"""
        return self._context

    @context.setter
    def context(self, context):
        """Set context
        Must be set AFTER self.non_user_depth
        """
        self._context = CONTEXTS[context]
        if self._context in (PACKAGE, ALL):
            dirname = os.path.dirname(self.path)
            for root, _, filenames in os.walk(dirname):
                for filename in fnmatch.filter(filenames, "*.py"):
                    self.add_path(os.path.join(root, filename))
        if context == ALL:
            self.non_user_depth = self.depth

    def clear_namespace(self, erase=True):
        """Clear namespace dict"""
        if erase:
            self.namespace.clear()
        self.namespace.update({
            "__name__": "__main__",
            "__file__": self.path,
            "__builtins__": __builtins__,
        })

    def clear_sys(self):
        """Clear sys variables"""
        # Replace now's dir with script's dir in front of module search path.
        sys.path[0] = self.dir
        # Clear argv
        sys.argv = self.argv

    def read_cmd_args(self, args, cmd=None):
        """Read cmd line argument object"""
        if not cmd:
            cmd = " ".join(sys.argv[1:])
        self.command = cmd
        self.dir = args.dir or os.path.dirname(args.script)
        self.argv = args.argv
        self.should_create_last_file = args.create_last
        self.name = args.name or os.path.basename(args.argv[0])
        self._read_args(args)
        self.path = args.script
        self.context = args.context
        return self

    def read_ipython_args(self, args, directory, filename, argv, create_last,   # pylint: disable=too-many-arguments
                          cmd=None):
        """Read magic line argument object"""
        if not cmd:
            cmd = "run " + " ".join(argv)
        self.command = cmd
        self.dir = directory or os.path.dirname(filename)
        self.argv = argv
        self.should_create_last_file = create_last
        self.name = args.name or os.path.basename(filename)
        self._read_args(args)
        self.path = filename
        self.context = args.context
        return self

    def _read_args(self, args):
        """Read cmd line argument object"""
        # ToDo #54: add serializer param
        self.serialize = get_serializer(args)
        self.verbose = args.verbose
        self.meta = args.meta

        self.disasm = args.disasm
        self.disasm0 = args.disasm0
        self.bypass_modules = args.bypass_modules

        self.depth = args.depth
        self.non_user_depth = args.non_user_depth
        self.execution_provenance = args.execution_provenance
        self.save_frequency = args.save_frequency
        self.call_storage_frequency = args.call_storage_frequency

        io.print_msg("setting up local provenance store")
        persistence_config.connect(self.dir)
        return self

    def read_restore_args(self, args):
        """Read cmd line argument object for 'now restore'"""
        self.bypass_modules = True
        self.command = " ".join(sys.argv[1:])

        self.dir = args.dir
        return self

    def create_last(self):
        """Create file indicating last trial id"""
        if self.should_create_last_file:
            lastname = os.path.join(os.path.dirname(self.path), LAST_TRIAL)
            with open(lastname, "w") as lastfile:
                lastfile.write(str(self.trial_id))

    def create_trial_args(self, args=None, run=True):
        """Return arguments for Trial.store"""
        if args is None:
            args = " ".join(sys.argv[1:])
        now = datetime.now()
        return (
            now, self.name, self.code_hash, args,
            self.bypass_modules, self.command, run,
            self.docstring
        )

    def create_automatic_tag_args(self):
        """Return arguments for Tag.create_automatic_tag"""
        return (self.trial_id, self.code_hash, self.command)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)
"""
Supporting infrastructure to run scientific experiments without a
scientific workflow management system.
"""
from noworkflow.now.cmd import main
from noworkflow.now.utils.functions import version


def load_ipython_extension(ipython):
    from .now.ipython import init
    init(ipython=ipython)


if __name__ == "__main__":
    main()

__version__ = version()



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Object Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer, Text
from sqlalchemy import PrimaryKeyConstraint, ForeignKeyConstraint
from sqlalchemy import CheckConstraint

from ...utils.prolog import PrologDescription, PrologTrial, PrologAttribute
from ...utils.prolog import PrologRepr

from .base import AlchemyProxy, proxy_class, backref_one


@proxy_class
class Object(AlchemyProxy):
    """Represent an object (global, argument, function call)"""

    __tablename__ = "object"
    __table_args__ = (
        PrimaryKeyConstraint("trial_id", "function_def_id", "id"),
        ForeignKeyConstraint(["trial_id"], ["trial.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id", "function_def_id"],
                             ["function_def.trial_id",
                              "function_def.id"], ondelete="CASCADE"),
    )
    trial_id = Column(Integer, index=True)
    function_def_id = Column(Integer, index=True)
    id = Column(Integer, index=True)                                             # pylint: disable=invalid-name
    name = Column(Text)
    type = Column(                                                               # pylint: disable=invalid-name
        Text,
        CheckConstraint("type IN ('GLOBAL', 'ARGUMENT', 'FUNCTION_CALL')"))

    trial = backref_one("trial")  # Trial.objects
    function_def = backref_one("function_def")  # FunctionDef.objects

    prolog_description = PrologDescription("object", (
        PrologTrial("trial_id", link="function_def.trial_id"),
        PrologAttribute("function_def_id", link="function_def.id"),
        PrologAttribute("id"),
        PrologRepr("name"),
        PrologRepr("type"),
    ), description=(
        "informs that in a given trial (*trial_id*),\n"
        "a given function definition (*function_def_id*),\n"
        "has a GLOBAL/ARGUMENT/FUNCTION_CALL (*type*),\n"
        "with *name*.\n"
    ))

    def __repr__(self):
        return (
            "Object({0.trial_id}, {0.function_def_id}, "
            "{0.id}, {0.name}, {0.type})"
        ).format(self)



def filter_five_out(value):
    if value == 5:
        return -1
    return value

def filter_even_out(value):
    if value % 2 == 0:
        return -1
    return value


with open("result.txt", "w") as result:
    for i in range(10):
        value = filter_even_out(i)
        if value == -1:
            continue
        value = filter_five_out(value)
        if value != -1:
            continue
        result.write(str(value))



# Dummy calls for representing openings

def align_warp(anatomy_image, anatomy_header, ref_image, ref_header):
    with open(anatomy_image, "rb"), open(anatomy_header, "rb"), \
            open(ref_image, "rb"), open(ref_header, "rb"), \
            open("warp.warp", "wb") as warp:
        if anatomy_image == "anatomy1.img":
            warp.write(b"\x01\x00\x00\x00\x04\x00\x00\x00")
        if anatomy_image == "anatomy2.img":
            warp.write(b"\x02\x00\x00\x00\x04\x00\x00\x00")
        if anatomy_image == "anatomy3.img":
            warp.write(b"\x03\x00\x00\x00\x04\x00\x00\x00")
        if anatomy_image == "anatomy4.img":
            warp.write(b"\x04\x00\x00\x00\x04\x00\x00\x00")
        return "warp.warp"

def reslice(warp, num=[1]):
    image = "reslice{}.img".format(num[0])
    header = "reslice{}.hdr".format(num[0])
    with open(warp, "rb") as w, \
            open(image, "w") as ri, open(header, "w") as rh:
        ri.write("image " + str(num[0]))
        rh.write("header " + str(num[0]))
    num[0] += 1
    return image, header

def softmean(rimag1, rhead1, rimag2, rhead2, rimag3, rhead3, rimag4, rhead4):
    with open(rimag1, "rb"), open(rhead1, "rb"), \
            open(rimag2, "rb"), open(rhead2, "rb"), \
            open(rimag3, "rb"), open(rhead3, "rb"), \
            open(rimag4, "rb"), open(rhead4, "rb"), \
            open("atlas.img", "w") as atlas_i, \
            open("atlas.hdr", "w") as atlas_h:
        atlas_i.write("atlas image")
        atlas_h.write("atlas header")
    return "atlas.img", "atlas.hdr"

def slicer(atlas_image, atlas_header, coordinate):
    output = "atlax-{}.pgm".format(coordinate)
    with open(atlas_image, "rb"), open(atlas_header, "rb"), \
            open(output, "w") as aslice:
        aslice.write("atlas {}".format(coordinate))
    return output

def convert(atlas_slice):
    result = atlas_slice[:-3] + "gif"
    with open(atlas_slice, "rb") as aslice, \
            open(result, "w") as gif:
        gif.write(atlas_slice)
    return result



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Execution Provider base"""

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
from collections import defaultdict

from .argument_captors import ArgumentCaptor


class ExecutionProvider(object):                                                 # pylint: disable=too-many-instance-attributes
    """Execution provider base class"""

    def __init__(self, metascript):
        # Indicates when activations should be collected
        #   (only after the first call to the script)
        self.enabled = False
        # User files
        self.script = metascript.path
        self.paths = metascript.paths
        # Which function types ('main', 'package' or 'all')
        #   should be considered for the threshold
        self.context = metascript.context
        # How deep we want to go when capturing function activations?
        self.depth_threshold = metascript.depth
        # How deep we want to go beyond our context
        self.non_user_depth_threshold = metascript.non_user_depth

        # Object serializer function
        self.serialize = metascript.serialize

        self.metascript = metascript
        self.trial_id = metascript.trial_id
        self.event_map = defaultdict(lambda: self.trace_empty, {})
        self.default_profile = sys.getprofile()
        self.default_trace = sys.gettrace()

        self.argument_captor = ArgumentCaptor(self)

    def trace_empty(self, frame, event, arg):
        """Call this function when trace event is not defined"""
        pass

    def store(self, partial=False):
        """Store provenance. Override it on subclasses"""
        pass

    def teardown(self):
        """Disable collection"""
        self.enabled = False

    def tearup(self):
        """Enable collection"""
        pass



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Graph Cache Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer, Text, TIMESTAMP

from .. import relational
from .base import AlchemyProxy, proxy_class, proxy_gen


@proxy_class
class GraphCache(AlchemyProxy):
    """Represent a Graph Cache"""

    __tablename__ = "graph_cache"
    __table_args__ = (
        {"sqlite_autoincrement": True},
    )
    id = Column(Integer, primary_key=True)                                       # pylint: disable=invalid-name
    type = Column(Text)                                                          # pylint: disable=invalid-name
    name = Column(Text)
    attributes = Column(Text)
    content_hash = Column(Text)
    duration = Column(Integer)
    timestamp = Column(TIMESTAMP)

    def __repr__(self):
        return "Cache({0.id}, {0.type}, {0.name})".format(self)

    @classmethod  # query
    def _query_select_cache(cls, gtype, name, attributes, session=None):
        """Find caches query by type, name and attributes
        Return sqlalchemy query
        """
        model = cls.m
        session = session or relational.session
        return session.query(model).filter(
            (model.type == gtype) &
            (model.name == name) &
            (model.attributes == attributes)
        )

    @classmethod  # query
    def select_cache(cls, gtype, name, attributes, session=None):
        """Find caches query by type, name and attributes


        Arguments:
        type -- cache type: trial, diff
        name -- graph mode: graph, tree, no_match, exact_match, namespace_match
        attributes -- other configuration: time


        Keyword arguments:
        session -- desired session
        """
        return proxy_gen(cls._query_select_cache(
            gtype, name, attributes, session=session or relational.session))

    @classmethod  # query
    def remove(cls, gtype, name, attributes, session=None, commit=False):        # pylint: disable=too-many-arguments
        """Remove caches query by type, name and attributes


        Arguments:
        type -- cache type: trial, diff
        name -- graph mode: graph, tree, no_match, exact_match, namespace_match
        attributes -- other configuration: time


        Keyword arguments:
        session -- desired session (default=relational.session)
        commit -- commit deletion (default=False)
        """
        cls._query_select_cache(
            gtype, name, attributes, session=session or relational.session
        ).delete()
        if commit:
            session.commit()

    @classmethod  # query
    def create(cls, gtype, name, duration, attributes, content_hash,             # pylint: disable=too-many-arguments
               session=None, commit=False):
        """Create Cache


        Arguments:
        type -- cache type: trial, diff
        name -- graph mode: graph, tree, no_match, exact_match, namespace_match
        duration -- required time to calculate graph
        attributes -- other configuration: time
        content_hash -- hash of stored graph


        Keyword arguments:
        session -- desired session (default=relational.session)
        commit -- commit insertion (default=False)
        """
        session = session or relational.session
        cache = cls.m(                                                           # pylint: disable=not-callable
            type=gtype, name=name, duration=duration,
            attributes=attributes, content_hash=content_hash
        )
        session.add(cache)
        if commit:
            session.commit()



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Module Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer, Text, select, bindparam

from .. import relational

from .base import AlchemyProxy, proxy_class, backref_many, is_none


@proxy_class
class Module(AlchemyProxy):
    """Represent a module"""

    __tablename__ = "module"
    __table_args__ = (
        {"sqlite_autoincrement": True},
    )
    id = Column(Integer, primary_key=True)                                       # pylint: disable=invalid-name
    name = Column(Text)
    version = Column(Text)
    path = Column(Text)
    code_hash = Column(Text, index=True)

    trials = backref_many("trials")  # Trial.dmodules

    @property
    def brief(self):
        """Brief description of module"""
        result = "{0.name}".format(self)
        if self.version:
            result += " {0.version}".format(self)
        return result

    def __key(self):
        return (self.name, self.version, self.path, self.code_hash)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()                                     # pylint: disable=protected-access

    def show(self, _print=print):
        """Show module"""
        _print("""\
            Name: {0.name}
            Version: {0.version}
            Path: {0.path}
            Code hash: {0.code_hash}\
            """.format(self))

    def __repr__(self):
        return "Module({0.id}, {0.name}, {0.version})".format(self)

    @classmethod  # query
    def id_seq(cls, session=None):
        """Load next module id

        Keyword arguments:
        session -- specify session for loading (default=relational.session)
        """
        session = session or relational.session
        an_id = session.execute(
            """SELECT seq
               FROM SQLITE_SEQUENCE
               WHERE name='module'"""
        ).fetchone()
        if an_id:
            return an_id[0]
        return 0

    @classmethod  # query
    def fast_load_module_id(cls, name, version, path, code_hash,                 # pylint: disable=too-many-arguments
                            session=None):
        """Load module id by name, version and code_hash

        Compile SQLAlchemy core query into string for optimization

        Keyword arguments:
        session -- specify session for loading (default=relational.session)
        """
        session = session or relational.session
        if not hasattr(cls, "_load_or_create_module_id"):
            tmodule = cls.t
            _query = select([tmodule.c.id]).where(
                (tmodule.c.name == bindparam("name")) &
                (tmodule.c.version == bindparam("version")) &
                (tmodule.c.code_hash == bindparam("code_hash"))
            )
            cls._load_or_create_module_id = str(_query)

        info = dict(name=name, path=path, version=version, code_hash=code_hash)
        an_id = session.execute(
            cls._load_or_create_module_id, info).fetchone()
        if an_id:
            return an_id[0]



import numpy as np
import matplotlib.pyplot as plt
from precipitation import read, prepare

def bar_graph(prec, months, years):
    prepare(prec, months, years, plt)
    plt.savefig("out.png")

months = np.arange(12) + 1
d13, d14 = read('p13.dat'), read('p14.dat')
prec = prec13, prec14 = [], []

for i in months:
    prec13.append(sum(d13[i]))
    prec14.append(sum(d14[i]))

bar_graph(prec, months, ['2013', '2014'])



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now history' command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from ..ipython.converter import create_ipynb
from ..persistence.models.history import History as HistoryModel
from ..persistence import persistence_config

from .command import NotebookCommand


class History(NotebookCommand):
    """Show project history"""

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg("-s", "--script", type=str, default="*",
                help="show history of specific script")
        add_arg("-e", "--status", type=str, default="*",
                choices=["*", "finished", "unfinished", "backup"],
                help="show only trials in a specific status")
        add_arg("--dir", type=str,
                help="set demo path. Default to CWD/demo<number>"
                     "where <number> is the demo identification")

    def execute(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())
        history = HistoryModel(script=args.script, status=args.status)
        print(history)

    def execute_export(self, args):
        code = ("%load_ext noworkflow\n"
                "import noworkflow.now.ipython as nip\n"
                "# <codecell>\n"
                "history = nip.History()\n"
                "# history.graph.width = 700\n"
                "# history.graph.height = 300\n"
                "# history.script = '*'\n"
                "# history.status = '*'\n"
                "# <codecell>\n"
                "history")
        create_ipynb("History.ipynb", code)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Magics for IPython"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from collections import defaultdict
from functools import wraps

from IPython.core.magic import Magics, magics_class

from .now_run import NowRun
from .now_ipython import NowIpython
from .now_set_default import NowSetDefault
from .now_sql import NowSQL
from .now_prolog import NowProlog
from .now_restore import NowRestore
from .now_schema import NowSchema
from .now_ls_magic import NowLsMagic


MAGICS = [
    ("now_run", "line_cell", NowRun),
    ("now_ip", "line", NowIpython),
    ("now_set_default", "line", NowSetDefault),
    ("now_sql", "cell", NowSQL),
    ("now_prolog", "cell", NowProlog),
    ("now_restore", "line", NowRestore),
    ("now_schema", "line", NowSchema),
    ("now_ls_magic", "line", NowLsMagic),
]


@magics_class
class NoworkflowMagics(Magics):
    """Generate noWorkflow magics"""

    def __init__(self, shell):
        super(NoworkflowMagics, self).__init__(shell=shell)
        self.commands = [
            cls(command, cls.__doc__, magic_type=magic_type)
            for command, magic_type, cls in MAGICS
        ]
        self.now_magics = defaultdict(dict)
        self._generate_magics()

    def _generate_magics(self):
        """Generate noWorkflow magics"""
        for command in self.commands:
            command.add_arguments()

            @wraps(command)
            def func(line, cell=None, command=command):
                """Magic"""
                return command.execute(command.func, line, cell, self)

            command.func = command.create_magic(func)
            for typ in command.magic_type.split("_"):
                self.magics[typ][command.magic] = command.func
                self.now_magics[typ][command.magic] = command.func


def register_magics(ipython):
    """Register IPython magics"""
    if not ipython:
        ipython = get_ipython()                                                  # pylint: disable=undefined-variable
    magics = NoworkflowMagics(ipython)
    ipython.register_magics(magics)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Diff Object"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from collections import OrderedDict

from future.utils import viewkeys

from .base import Model, proxy_gen
from .graphs.diff_graph import DiffGraph
from .trial import Trial


class Diff(Model):
    """This model represents a diff between two trials
    Initialize it by passing both trials ids:
        diff = Diff(2)

    There are four visualization modes for the graph:
        tree: activation tree without any filters
            diff.graph.mode = 0
        no match: tree transformed into a graph by the addition of sequence and
                  return edges and removal of intermediate call edges
            diff.graph.mode = 1
        exact match: calls are only combined when all the sub-call match
            diff.graph.mode = 2
        namesapce: calls are combined without considering the sub-calls
            diff.graph.mode = 3


    There are also three visualization modes for the diff:
        combine graphs: combines both trial graphs
            diff.graph.view = 0
        side by side: displays both graphs side by side
            diff.graph.view = 1
        combined and side by side: combine graphs and displays both separated
            diff.graph.view = 2


    You can change the graph width and height by the variables:
        diff.graph.width = 600
        diff.graph.height = 400
    """

    __modelname__ = "Diff"

    DEFAULT = {
        "graph.width": 500,
        "graph.height": 500,
        "graph.mode": 3,
        "graph.view": 0,
        "graph.neighborhoods": 3,
        "graph.time_limit": None,
    }

    REPLACE = {
        "graph_width": "graph.width",
        "graph_height": "graph.height",
        "graph_mode": "graph.mode",
        "graph_view": "graph.view",
        "graph_neighborhoods": "graph.neighborhoods",
        "graph_time_limit": "graph.time_limit",
    }

    def __init__(self, trial_ref1, trial_ref2, **kwargs):
        super(Diff, self).__init__(trial_ref1, trial_ref2, **kwargs)
        self.trial1 = Trial(trial_ref1)
        self.trial2 = Trial(trial_ref2)

        self.graph = DiffGraph(self)
        self.initialize_default(kwargs)

    @property
    def trial(self):
        """Return a tuple with information from both trials """
        extra = ("start", "finish", "duration_text")
        ignore = ("id",)
        return diff_dict(
            self.trial1.to_dict(ignore=ignore, extra=extra),                     # pylint: disable=no-member
            self.trial2.to_dict(ignore=ignore, extra=extra))                     # pylint: disable=no-member

    @property
    def modules(self):
        """Diff modules from trials"""
        return diff_set(
            set(proxy_gen(self.trial1.modules)),
            set(proxy_gen(self.trial2.modules)))

    @property
    def environment(self):
        """Diff environment variables"""
        return diff_set(
            set(self.trial1.environment_attrs),
            set(self.trial2.environment_attrs))

    @property
    def file_accesses(self):
        """Diff file accesses"""
        return diff_set(
            set(self.trial1.file_accesses),
            set(self.trial2.file_accesses),
            create_replaced=False)

    def _repr_html_(self):
        return self.graph._repr_html_()                                          # pylint: disable=protected-access


def diff_dict(before, after):
    """Compare dicts.
    Return a dict with keys shared by both dicts that have different values
        key -> [before[key], after[key]]
    """
    result = OrderedDict()
    for key in viewkeys(before):
        if key != "id" and before[key] != after[key]:
            result[key] = [before[key], after[key]]
    return result


def diff_set(before, after, create_replaced=True):
    """Compare sets to get additions, removals and replacements

    Return 3 sets:
    added -- objects present in second set, but not present in first set
    removed -- objects present in first set, but not present in second set
    replaced -- objects that have the same name in both sets, but are different
    """
    removed = before - after
    added = after - before
    replaced = set()

    removed_by_name = {}
    for element_removed in removed:
        removed_by_name[element_removed.name] = element_removed
    for element_added in added:
        element_removed = removed_by_name.get(element_added.name)
        if element_removed and create_replaced:
            replaced.add((element_removed, element_added))
    if create_replaced:
        for (element_removed, element_added) in replaced:
            removed.discard(element_removed)
            added.discard(element_added)

    return (added, removed, replaced)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Intermediate Tree Structures and Graph Structures"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


import json
import time
import traceback

from collections import OrderedDict

from future.utils import viewitems
from sqlalchemy import exc

from ....persistence import relational, content
from ....utils.cross_version import pickle
from ....utils.data import OrderedCounter
from ....utils.io import print_msg

from .. import GraphCache


class Graph(object):                                                             # pylint: disable=too-few-public-methods
    """Graph superclass. Handle json transformation"""
    def escape_json(self, data):                                                 # pylint: disable=no-self-use
        """Escape JSON"""
        data = json.dumps(data)
        return (data.replace("&", "\\u0026")
                .replace("<", "\\u003c")
                .replace(">", "\\u003e"))


def prepare_cache(get_type):
    """Decorator: Load graph from cache"""
    def cache(name, attrs=""):
        """Decorator: Load graph from cache"""
        def dec(func):
            """Decorator: Load graph from cache"""
            def load(self, *args, **kwargs):
                """Load graph from cache

                Find graph by type, name and attributes
                If graph is cached, return it

                Return:
                finished -- trial has finished
                graph -- cached trial graph
                """
                cache_session = relational.make_session()

                typ = get_type(self, *args, **kwargs)
                attributes = " ".join(str(kwargs[a])
                                      for a in attrs.split() if a in kwargs)

                information = (typ, name, attributes)
                if self.use_cache:
                    try:
                        caches = GraphCache.select_cache(*information,
                                                         session=cache_session)
                        for cache in caches:
                            result = pickle.loads(
                                content.get(cache.content_hash))
                            if not result[0]:
                                continue
                            cache_session.close()                                # pylint: disable=no-member
                            return result
                    except (ValueError, exc.SQLAlchemyError):
                        traceback.print_exc()
                        print_msg("Couldn't load graph cache", True)
                start = time.time()
                graph = func(self, *args, **kwargs)
                duration = time.time() - start
                try:
                    GraphCache.remove(*information, session=cache_session)
                    GraphCache.create(
                        typ, name, duration, attributes,
                        content.put(pickle.dumps(graph)),
                        session=cache_session, commit=True
                    )
                except exc.SQLAlchemyError:
                    traceback.print_exc()
                    print_msg("Couldn't store graph cache", True)
                cache_session.close()                                            # pylint: disable=no-member
                return graph
            return load
        return dec
    return cache


class TreeElement(object):
    """Base class for Intermediate Tree"""

    def __init__(self, level=-1, use_id=True):
        self.duration = 0
        self.count = 1
        self.repr = ""
        self.level = level
        self.use_id = use_id
        self.trial_id = 0

    def mean(self):
        """Mean duration of Tree Node"""
        return self.duration / self.count

    def visit(self, visitor):
        """Visitor pattern"""
        name = "visit_{}".format(self.__class__.__name__.lower())
        return getattr(visitor, name)(self)

    def calculate_repr(self):
        """Calculate Representation of Node"""
        pass

    def mix(self, other):
        """Combine Nodes"""
        pass

    def __hash__(self):
        if self.use_id:
            return id(self)
        return hash(self.__repr__())

    def __repr__(self):
        return self.repr

    def to_dict(self, nid):                                                      # pylint: disable=unused-argument, no-self-use
        """Convert element to dict"""
        return {}


class Single(TreeElement):                                                       # pylint: disable=too-many-instance-attributes
    """Single Node
       Represent an activation or a group of merged activations"""
    def __init__(self, activation):
        super(Single, self).__init__(level=0, use_id=True)
        self.activation = activation
        self.activations = {activation}
        self.parent = activation.caller_id
        self.id = activation.id                                                  # pylint: disable=invalid-name
        self.line = activation.line

        name = activation.name
        if activation.name == "_handle_fromlist":
            for argument in activation.arguments:
                if argument.name == "module":
                    name = (
                        argument.value
                        .split("' from '")[-1]
                        .split('" from "')[-1]
                        [:-2]
                    )
        self.name = name
        self.trial_id = activation.trial_id
        self.repr = "S({0}-{1})".format(self.line, self.name)
        self.finished = bool(activation.finish)

    @property
    def count(self):
        """Count activations"""
        return sum(1 for a in self.activations)

    @count.setter
    def count(self, value):
        """Ignore set count"""
        pass

    @property
    def duration(self):
        """Calculate total duration"""
        return sum(a.duration for a in self.activations
                   if a.finish and a.start)

    @duration.setter
    def duration(self, value):
        """Ignore set duration"""
        pass

    def mix(self, other):
        """Combine activations"""
        self.finished &= other.finished
        self.count += other.count
        self.duration += other.duration
        self.activations = self.activations.union(other.activations)

    def __eq__(self, other):
        return all(fn(self) == fn(other) for fn in
                   [type, lambda x: x.line, lambda x: x.name])

    def name_id(self):
        """Return name that identifies node"""
        return "{0} {1}".format(self.line, self.name)

    def visit(self, visitor):
        return visitor.visit_single(self)

    def to_dict(self, nid):
        return {
            "index": nid,
            "caller_id": self.parent,
            "name": self.name,
            "node": {
                "trial_id": self.trial_id,
                "line": self.line,
                "count": self.count,
                "duration": self.duration,
                "level": self.level,
                "finished": self.finished,
                "info": Info(self),
            },
        }

    def __hash__(self):
        return super(Single, self).__hash__()


class Mixed(TreeElement):
    """Mixed Node
       Represent complex Mixed Node"""
    def __init__(self, activation):
        super(Mixed, self).__init__(level=0, use_id=True)
        self.duration = activation.duration
        self.elements = [activation]
        self.parent = activation.parent
        self.id = activation.id                                                  # pylint: disable=invalid-name
        self.repr = activation.repr
        self.finished = activation.finished

    @property
    def count(self):
        """Count activations"""
        return sum(e.count for e in self.elements)

    @count.setter
    def count(self, value):
        """Ignore set count"""
        pass

    @property
    def duration(self):
        """Calculate total duration"""
        return sum(e.duration for e in self.elements)

    @duration.setter
    def duration(self, value):
        """Ignore set duration"""
        pass

    @property
    def first(self):
        """Get first node"""
        return next(iter(self.elements))

    def add_element(self, element):
        """Add node"""
        self.finished &= element.finished
        self.elements.append(element)

    def mix(self, other):
        """Add nodes"""
        for element in other.elements:
            self.finished &= element.finished

        self.elements += other.elements
        self.mix_results()

    def mix_results(self):
        """Combine results"""
        iterator = iter(self.elements)
        initial = next(iterator)
        for element in iterator:
            initial.mix(element)

    def __hash__(self):
        return super(Mixed, self).__hash__()


class Group(TreeElement):                                                        # pylint: disable=too-many-instance-attributes
    """Group Node
       Represent a group of called nodes"""
    def __init__(self):
        super(Group, self).__init__(level=0, use_id=True)
        self.nodes = OrderedDict()
        self.edges = OrderedDict()
        self.duration = 0
        self.parent = None
        self.count = 1
        self.repr = ""
        self.finished = True
        self.next_element = None
        self.last = None

    def initialize(self, previous, nnext):
        """Initialize the group node by two nodes"""
        self.nodes[nnext] = Mixed(nnext)
        self.duration = nnext.duration
        self.next_element = nnext
        self.last = nnext
        self.add_subelement(previous)
        self.parent = nnext.parent
        self.finished = previous.finished and nnext.finished
        return self

    def add_subelement(self, previous):
        """Add new node in sequence to group"""
        self.finished &= previous.finished
        nnext, self.next_element = self.next_element, previous
        if previous not in self.edges:
            self.edges[previous] = OrderedCounter()
        if previous not in self.nodes:
            self.nodes[previous] = Mixed(previous)
        else:
            self.nodes[previous].add_element(previous)
        self.edges[previous][nnext] += 1

    def calculate_repr(self):
        result = [
            "[{0}-{1}->{2}]".format(previous, count, nnext)
            for previous, edges in viewitems(self.edges)
            for nnext, count in viewitems(edges)
        ]

        self.repr = "G({0})".format(", ".join(result))

    def __eq__(self, other):
        return all(fn(self) == fn(other) for fn in
                   [type, lambda x: x.edges])

    def mix(self, other):
        for node, value in viewitems(self.nodes):
            value.mix(other.nodes[node])

    def __hash__(self):
        return super(Group, self).__hash__()


class Call(TreeElement):                                                         # pylint: disable=too-many-instance-attributes
    """Call Node
       Represent a call from a Node to a Group"""
    def __init__(self, caller, called):
        super(Call, self).__init__(level=0, use_id=True)
        self.caller = caller
        self.called = called
        self.called.calculate_repr()
        self.parent = caller.parent
        self.count = 1
        self.id = self.caller.id                                                 # pylint: disable=invalid-name
        self.duration = self.caller.duration
        self.repr = "C({0}, {1})".format(self.caller, self.called)
        self.finished = self.caller.finished

    def __eq__(self, other):
        return all(fn(self) == fn(other) for fn in
                   [type, lambda x: x.caller, lambda x: x.called])

    def mix(self, other):
        self.caller.mix(other.caller)
        self.called.mix(other.called)

    def __hash__(self):
        return super(Call, self).__hash__()


def duration_text(duration, count):
    """Return duration text"""
    return "Total duration: {} microseconds for {} activations".format(
        duration, count)


def mean_text(mean):
    """Return mean text"""
    return "Mean: {} microseconds per activation".format(mean)


def activation_text(activation):
    """Return single activation text"""
    extra = "(still running)"
    if activation.finish:
        extra = "to {a.finish} ({d} microseconds)".format(
            d=activation.duration, a=activation)
    result = [
        "",
        "Activation #{a.id} from {a.start} {extra}".format(
            extra=extra, a=activation),
    ]

    activation.show(
        _print=lambda x, offset=0: result.append(offset * "&nbsp;" + x))

    return result


class Info(object):
    """Info class
       Represent the information of a node"""

    def __init__(self, single):
        self.title = (
            "Trial {trial}<br>"
            "Function <b>{name}</b> called at line {line}").format(
                trial=single.trial_id, name=single.name, line=single.line)
        self.activations = set()
        self.duration = ""
        self.mean = ""
        self.extract_activations(single)
        self.activation_list = None

    def update_by_node(self, node):
        """Update information by node"""
        self.duration = duration_text(node["duration"], node["count"])
        self.mean = mean_text(node["mean"])
        self.activation_list = sorted(self.activations, key=lambda a: a[0])

    def add_activation(self, activation):
        """Add activation information"""
        self.activations.add((activation.start, activation))

    def extract_activations(self, single):
        """Add activations information"""
        for activation in single.activations:
            self.add_activation(activation)

    def __repr__(self):
        result = [self.title, self.duration, self.mean]
        for activation in self.activation_list:
            result += activation_text(activation[1])

        return "<br/>".join(result)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Module Dependency Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer
from sqlalchemy import ForeignKeyConstraint, PrimaryKeyConstraint

from ...utils.prolog import PrologDescription, PrologTrial, PrologAttribute
from ...utils.prolog import PrologRepr, PrologNullableRepr

from .base import AlchemyProxy, proxy_class, one, backref_one


@proxy_class
class Dependency(AlchemyProxy):
    """Dependency proxy

    Use it to have different objects with the same primary keys
    Use it also for re-attaching objects to SQLAlchemy (e.g. for cache)
    """

    __tablename__ = "dependency"
    __table_args__ = (
        PrimaryKeyConstraint("trial_id", "module_id"),
        ForeignKeyConstraint(["trial_id"], ["trial.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["module_id"], ["module.id"], ondelete="CASCADE"),
    )
    trial_id = Column(Integer, nullable=False, index=True)
    module_id = Column(Integer, nullable=False, index=True)

    module = one("Module")

    trial = backref_one("trial")  # Trial.module_dependencies

    prolog_description = PrologDescription("module", (
        PrologTrial("trial_id", link="trial.id"),
        PrologAttribute("id", attr_name="module.id"),
        PrologRepr("name", attr_name="module.name"),
        PrologNullableRepr("version", attr_name="module.version"),
        PrologNullableRepr("path", attr_name="module.path"),
        PrologNullableRepr("code_hash", attr_name="module.code_hash"),
    ), description=(
        "informs that a given trial (*trial_id*)\n"
        "imported the *version* of a module (*name*),\n"
        "with content (*code_hash*) written in *path*."
    ))

    def __repr__(self):
        return "Dependency({0.trial_id}, {0.module})".format(self)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Commands and argument parsers for 'now'"""
from __future__ import (absolute_import, print_function,
                        division)

import argparse
import sys
import sqlalchemy

from .command import Command, SmartFormatter
from .cmd_run import Run
from .cmd_debug import Debug
from .cmd_list import List
from .cmd_show import Show
from .cmd_diff import Diff
from .cmd_dataflow import Dataflow
from .cmd_export import Export
from .cmd_restore import Restore
from .cmd_vis import Vis
from .cmd_demo import Demo
from .cmd_history import History
from .cmd_schema import Schema
from ..utils.io import print_msg


def main():
    """Main function"""
    from ..utils.functions import version
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=SmartFormatter)
    parser.add_argument("-v", "--version", action="version",
                        version="noWorkflow {}".format(version()))
    subparsers = parser.add_subparsers()
    commands = [
        Run(),
        Debug(),
        List(),
        Show(),
        Diff(),
        Dataflow(),
        Export(),
        Restore(),
        Vis(),
        Demo(),
        History(),
        Schema(),
    ]
    for cmd in commands:
        cmd.create_parser(subparsers)

    if len(sys.argv) == 1:
        sys.argv.append("-h")

    try:
        args, _ = parser.parse_known_args()
        args.func(args)
    except RuntimeError as exc:
        print_msg(exc, True)
    except sqlalchemy.exc.OperationalError as exc:
        print_msg("invalid noWorkflow database", True)
        print_msg("it is probably outdated", True)


__all__ = [
    "Command",
    "Run",
    "Debug",
    "List",
    "Show",
    "Diff",
    "Dataflow",
    "Export",
    "Restore",
    "Vis",
    "Demo",
    "History",
    "main",
]



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Slicing Usage Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer, Text
from sqlalchemy import PrimaryKeyConstraint, ForeignKeyConstraint
from sqlalchemy import CheckConstraint

from ...utils.prolog import PrologDescription, PrologTrial, PrologAttribute
from ...utils.prolog import PrologRepr

from .base import AlchemyProxy, proxy_class, backref_one


@proxy_class
class VariableUsage(AlchemyProxy):
    """Represent a variable usage during program slicing"""

    __tablename__ = "variable_usage"
    __table_args__ = (
        PrimaryKeyConstraint("trial_id", "id"),
        ForeignKeyConstraint(["trial_id"],
                             ["trial.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id", "activation_id"],
                             ["function_activation.trial_id",
                              "function_activation.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id", "activation_id", "variable_id"],
                             ["variable.trial_id",
                              "variable.activation_id",
                              "variable.id"], ondelete="CASCADE"),
    )
    trial_id = Column(Integer, index=True)
    activation_id = Column(Integer, index=True)
    variable_id = Column(Integer, index=True)
    id = Column(Integer, index=True)                                             # pylint: disable=invalid-name
    line = Column(Integer)
    context = Column(Text, CheckConstraint("context IN ('Load', 'Del')"))

    trial = backref_one("trial")  # Trial.variable_usages
    activation = backref_one("activation")  # Activation.variables_usages
    variable = backref_one("variable")  # Variable.usages

    prolog_description = PrologDescription("usage", (
        PrologTrial("trial_id", link="variable.trial_id"),
        PrologAttribute("activation_id", link="variable.activation_id"),
        PrologAttribute("variable_id", link="variable.id"),
        PrologAttribute("id"),
        PrologRepr("name", attr_name="variable.name"),
        PrologAttribute("line"),
    ), description=(
        "informs that in a given trial (*trial_id*),\n"
        "during a specific function activation (*activation_id*),\n"
        "in a specific *line* of code,\n"
        "a variable *name* was accessed (read, delete)."
    ))

    def __repr__(self):
        name = self.variable.name
        return (
            "VariableUsage({self.trial_id}, {self.activation_id}, "
            "{self.variable_id}, {self.id}, {name}, {self.line}, "
            "{self.context})"
        ).format(**locals())

    def __str__(self):
        name = self.variable.name
        return "(L{self.line}, {name}, <{self.context}>)".format(**locals())



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define utility functions"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import inspect

from os.path import join, dirname
from textwrap import dedent
from pkg_resources import resource_string, resource_listdir, resource_isdir


MODULE = __name__
MODULE = MODULE[:MODULE.rfind(".")]
MODULE = MODULE[:MODULE.rfind(".")]
NOWORKFLOW_DIR = dirname(dirname(dirname(__file__)))


def wrap(string, initial="  ", other="\n  "):
    """Re-indent indented text"""
    return initial + other.join(dedent(string).split("\n"))


def resource(filename, encoding=None):
    """Access resource content via setuptools"""
    content = resource_string(MODULE, filename)
    if encoding:
        return content.decode(encoding=encoding)
    return content


def resource_ls(path):
    """Access resource directory via setuptools"""
    return resource_listdir(MODULE, path)


def resource_is_dir(path):
    """Access resource directory via setuptools"""
    return resource_isdir(MODULE, path)


def version():
    """Return noWorkflow version"""
    return resource("../resources/version.txt", encoding="utf-8").strip()


def abstract():
    """Raise abstract Exception"""
    frame = inspect.currentframe().f_back
    name = frame.f_code.co_name
    raise Exception("Abstract method: {}".format(name))



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define bytecode interpreter that supports iteration on bytecode"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import dis
from dis import opmap

from collections import defaultdict
from opcode import HAVE_ARGUMENT, cmp_op

from .instructions import Instruction


CALL_FUNCTIONS = {
    opmap["CALL_FUNCTION"], opmap["CALL_FUNCTION_VAR"],
    opmap["CALL_FUNCTION_KW"], opmap["CALL_FUNCTION_VAR_KW"]
}

PRINT_ITEMS = set()
if "PRINT_ITEM" in opmap:
    PRINT_ITEMS.add(opmap["PRINT_ITEM"])
    PRINT_ITEMS.add(opmap["PRINT_ITEM_TO"])

PRINT_NEW_LINES = set()
if "PRINT_NEWLINE" in opmap:
    PRINT_NEW_LINES.add(opmap["PRINT_NEWLINE"])
    PRINT_NEW_LINES.add(opmap["PRINT_NEWLINE_TO"])

SETUP_WITH = {opmap["SETUP_WITH"], }
WITH_CLEANUP = {opmap.get("WITH_CLEANUP") or opmap.get("WITH_CLEANUP_START"), }
SETUP_ASYNC_WITH = set()
if "SETUP_ASYNC_WITH" in opmap:
    SETUP_ASYNC_WITH.add(opmap["SETUP_ASYNC_WITH"])

IMPORTS = {opmap["IMPORT_NAME"], opmap["IMPORT_FROM"]}

IMPORT_NAMES = {opmap["IMPORT_NAME"],}

FOR_ITERS = {opmap["FOR_ITER"],}
GET_ITERS = {opmap["GET_ITER"],}


def cord(value):
    """Convert (str or int) to ord"""
    if isinstance(value, str):
        return ord(value)
    return value


class ListAccessor(object):                                                      # pylint: disable=too-few-public-methods
    """List Proxy. Return value on x[i] and tuple on x(i)"""

    def __init__(self, values, repr_is_val=True):
        self.values = values
        self.repr_is_val = repr_is_val

    def __getitem__(self, index):
        if self.values is not None:
            return self.values[index]
        return index

    def __call__(self, index):
        argval = self[index]
        if self.repr_is_val and self.values is not None:
            argrepr = argval
        else:
            argrepr = repr(argval)
        return argval, argrepr


class Interpreter(object):                                                       # pylint: disable=too-many-instance-attributes
    """Bytecode iterator"""

    def __init__(self, co_code, varnames=None, names=None, constants=None,       # pylint: disable=too-many-arguments
                 cells=None, linestarts=None, line_offset=0):
        self.lasti = 0
        self.opi = 0
        self._extended_arg = 0

        self._co_code = co_code
        self.varnames = ListAccessor(varnames)
        self.names = ListAccessor(names)
        self.consts = ListAccessor(constants, repr_is_val=False)
        self.cells = ListAccessor(cells)
        self.linestarts = linestarts
        self.line_offset = line_offset

        self._size = len(co_code)

        self.opcode = None
        self.oparg = 0

        self._stop = False

        self._map = {}
        self._extra = set()
        self._missing = set()
        self._supported = set()
        if not hasattr(self, "_known_missing"):
            self._known_missing = set()

        self._create_map()

    def __iter__(self):
        """Restart iterator"""
        self._stop = False
        return self

    def __call__(self, lasti=0, extended_arg=0):
        self.lasti = lasti
        self._extended_arg = extended_arg

    def next(self):
        """Python 2 iterator"""
        if self._stop:
            raise StopIteration
        opcode = self._next_op()
        self._map[opcode]()
        return opcode

    def __next__(self):
        """Python 3 iterator"""
        return self.next()

    def _next_op(self):
        """Get next operation"""
        self._set_opcode()
        if self.opcode >= HAVE_ARGUMENT:
            self._have_argument()

        if self.lasti >= self._size:
            self._stop = True

        return self.opcode

    def _set_opcode(self):
        """Get op from code"""
        self.oparg = None
        self.opcode = cord(self._co_code[self.lasti])
        self.opi = self.lasti
        self.lasti += 1

    def _have_argument(self):
        """Read argument if op has argument"""
        cod = self._co_code
        i = self.lasti
        self.oparg = cord(cod[i]) + cord(cod[i + 1]) * 256 + self._extended_arg
        self._extended_arg = 0
        self.lasti += 2

    def _create_map(self):
        """Create map of functions"""
        condition = lambda x, obj: (
            x[0] != "_" and hasattr(obj, "__call__") and
            obj.__doc__ is not None and "opcode" in obj.__doc__)
        to_opcode = lambda x: x.upper().replace("__", "+")

        self._map = defaultdict(lambda: self.nop)
        self._extra = set()
        self._missing = set()
        self._supported = set()
        for name in dir(self):
            method = getattr(self, name)
            if condition(name, method):
                opcode = to_opcode(name)
                if opcode not in opmap:
                    self._extra.add(opcode)
                else:
                    self._map[opmap[opcode]] = method
                    self._supported.add(opcode)
        self._missing = (
            set(opmap.keys()) - self._supported - self._known_missing)

    @property
    def extra_opcode(self):
        """Return opcode implemented by this class
        but not supported by Python
        """
        return self._extra

    @property
    def missing_opcode(self):
        """Return opcode supported by Python
        but not implemented by this class"""
        return self._missing

    def nop(self):
        """NOP opcode"""
        pass


class InstructionInterpreter(Interpreter):
    """Mix Python3 dis._get_instructions_bytes with Python2 dis.disassemble"""

    def __init__(self, *args, **kwargs):
        super(InstructionInterpreter, self).__init__(*args, **kwargs)

        self._labels = dis.findlabels(self._co_code)

        self.starts_line = None
        self.is_jump_target = False
        self.argval = None
        self.argrepr = None
        self.current_line = -1

    def _set_opcode(self):
        super(InstructionInterpreter, self)._set_opcode()
        if self.linestarts is not None:
            self.starts_line = self.linestarts.get(self.opi, None)
            if self.starts_line is not None:
                self.starts_line += self.line_offset
                self.current_line = self.starts_line
        self.is_jump_target = self.opi in self._labels

    def _have_argument(self):
        super(InstructionInterpreter, self)._have_argument()
        opcode = self.opcode
        arg = argval = self.oparg
        argrepr = ""
        if opcode in dis.hasconst:
            argval, argrepr = self.consts(arg)
        elif opcode in dis.hasname:
            argval, argrepr = self.names(arg)
        elif opcode in dis.hasjrel:
            argval = self.lasti + arg
            argrepr = "to " + repr(argval)
        elif opcode in dis.haslocal:
            argval, argrepr = self.varnames(arg)
        elif opcode in dis.hascompare:
            argval = cmp_op[arg]
            argrepr = argval
        elif opcode in dis.hasfree:
            argval, argrepr = self.cells(arg)
        elif opcode in CALL_FUNCTIONS:
            argrepr = "%d positional, %d keyword pair" % (
                cord(self._co_code[self.lasti - 2]),
                cord(self._co_code[self.lasti - 1]))
        self.argval, self.argrepr = argval, argrepr

    def next(self):
        super(InstructionInterpreter, self).next()
        return Instruction(
            dis.opname[self.opcode], self.opcode, self.oparg, self. argval,
            self.argrepr, self.opi, self.starts_line, self.is_jump_target,
            self.current_line)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'%now_ls_magic' magic"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from ...utils.formatter import PrettyLines
from .command import IpythonCommandMagic


class NowLsMagic(IpythonCommandMagic):
    """Return the list of noWorkflow magics"""

    def execute(self, func, line, cell, magic_cls):
        cline, ccell = "%", "%%"
        magics = magic_cls.now_magics
        out = ["Line magics:",
               cline + ("  " + cline).join(sorted(magics["line"])),
               "",
               "Cell magics:",
               ccell + ("  " + ccell).join(sorted(magics["cell"]))]
        return PrettyLines(out)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""AST Visitors to capture definition provenance for slicing"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import ast
import sys

from collections import defaultdict, namedtuple

from ...utils.bytecode.interpreter import CALL_FUNCTIONS, PRINT_ITEMS
from ...utils.bytecode.interpreter import PRINT_NEW_LINES, SETUP_WITH
from ...utils.bytecode.interpreter import WITH_CLEANUP, SETUP_ASYNC_WITH
from ...utils.bytecode.interpreter import IMPORT_NAMES
from ...utils.bytecode.interpreter import FOR_ITERS, GET_ITERS

from .function_visitor import FunctionVisitor
from .utils import NamedContext, FunctionCall, ClassDef, Decorator, Generator
from .utils import Assert, safeget, With, Print, Import, ForIter, GeneratorCall
from .utils import Dependency, variable, Loop, Condition


class AssignLeftVisitor(ast.NodeVisitor):
    """Visit the left side of assignements and collect names"""
    def __init__(self):
        self.names = []
        self.enable = True
        self.last = ""

    def visit_Attribute(self, node):                                             # pylint: disable=invalid-name
        """Create nested names if it is an attribute"""
        self.generic_visit(node)
        if self.enable:
            self.last += "." + node.attr
            if not isinstance(self.last, str):
                self.last = self.last.encode("utf-8")
            self.names.append((self.last, node.ctx, node.lineno))

    def visit_Subscript(self, node):                                             # pylint: disable=invalid-name
        """Disable visitor for slice"""
        self.visit(node.value)
        self.enable = False
        self.visit(node.slice)

    def visit_Name(self, node):                                                  # pylint: disable=invalid-name
        """Collect names"""
        if self.enable:
            self.last = node.id
            self.names.append((self.last, node.ctx, node.lineno))
        self.generic_visit(node)


class AssignRightVisitor(ast.NodeVisitor):
    """Visit the right side of assignements and collect names"""

    def __init__(self):
        self.names = []
        self.special = NamedContext()
        self.line = -1

    def add(self, name, ctx, lineno):
        """Add name to current context"""
        if not self.special.use:
            self.names.append((name, ctx, lineno))
        else:
            self.special.add(name)

    def in_special(self, node):
        """Check if node is in special list for function calls"""
        return node.id in self.special.flat()

    def max_line(self, node):
        """Update the line according to the max lineno found"""
        try:
            self.line = max(node.lineno, self.line)
        except (TypeError, AttributeError):
            pass

    def new_comprehension(self, node):
        """Add call to comprehension"""
        self.max_line(node)
        self.add(variable(node.uid, "call"), "fn", node.lineno)

    def visit_Name(self, node):                                                  # pylint: disable=invalid-name
        """Collect names"""
        self.max_line(node)
        if not self.in_special(node):
            self.add(node.id, node.ctx, node.lineno)
        self.generic_visit(node)

    def visit_Lambda(self, node):                                                # pylint: disable=invalid-name
        """Create special context for lambda"""
        self.max_line(node)
        self.special.enable()
        self.visit(node.args)
        self.special.disable()
        self.visit(node.body)
        self.special.pop()

    def _visit_ListComp(self, node):                                             # pylint: disable=invalid-name
        """Create special context for ListComp"""
        self.max_line(node)
        #self.add(node.uid, "fn", node.lineno)
        self.special.enable()
        self.special.disable()
        for gen in node.generators:
            self.visit(gen)
        self.visit(node.elt)
        self.special.pop()

    def visit_ListComp(self, node):                                              # pylint: disable=invalid-name
        """Visit LiComp"""
        self.new_comprehension(node)

    if sys.version_info < (3, 0):
        visit_ListComp = _visit_ListComp                                         # pylint: disable=invalid-name

    def visit_SetComp(self, node):                                               # pylint: disable=invalid-name
        """Visit SetComp"""
        self.new_comprehension(node)

    def visit_GeneratorExp(self, node):                                          # pylint: disable=invalid-name
        """Visit GeneratorExp"""
        self.new_comprehension(node)

    def visit_DictComp(self, node):                                              # pylint: disable=invalid-name
        """Create special context for DictComp"""
        self.new_comprehension(node)

    def visit_comprehension(self, node):
        """Create special context for comprehension"""
        self.max_line(node)
        self.special.use = True
        self.visit(node.target)
        self.special.disable()
        self.visit(node.iter)
        for _if in node.ifs:
            self.visit(_if)

    def visit_Call(self, node):                                                  # pylint: disable=invalid-name
        """Create special dependency for calls"""
        self.max_line(node)
        self.add(variable(node.uid, "call"), "fn", node.lineno)

    def visit_Print(self, node):                                                 # pylint: disable=invalid-name
        """Create special dependency for calls"""
        self.max_line(node)
        self.add(variable(node.uid, "print"), "fn print", node.lineno)

    def visit_Import(self, node):                                                # pylint: disable=invalid-name
        """Create special dependency for imports"""
        self.max_line(node)
        self.add(variable(node.uid, "import"), "import", node.lineno)

    def visit_ImportFrom(self, node):                                            # pylint: disable=invalid-name
        """Create special dependency for imports"""
        self.max_line(node)
        self.add(variable(node.uid, "import from"), "import from", node.lineno)


class ComprehensionVisitor(AssignRightVisitor):
    """Dependency of variables in comprehension"""

    def visit_ListComp(self, node):
        """Visit ListComp"""
        self._visit_ListComp(node)

    def visit_SetComp(self, node):                                               # pylint: disable=invalid-name
        """Visit SetComp"""
        self._visit_ListComp(node)

    def visit_GeneratorExp(self, node):                                          # pylint: disable=invalid-name
        """Visit GeneratorExp"""
        self._visit_ListComp(node)

    def visit_DictComp(self, node):                                              # pylint: disable=invalid-name
        """Create special context for DictComp"""
        self.max_line(node)
        self.special.enable()
        self.special.disable()
        for gen in node.generators:
            self.visit(gen)
        self.visit(node.key)
        self.visit(node.value)
        self.special.pop()

    def visit_Call(self, node):                                                  # pylint: disable=invalid-name
        """Create special dependency for calls"""
        self.max_line(node)
        self.generic_visit(node)


def tuple_or_list(node):
    """Check if node is tuple or list"""
    return isinstance(node, ast.Tuple) or isinstance(node, ast.List)


def assign_dependencies(target, value, dependencies, typ,                        # pylint: disable=too-many-arguments
                        dep_typ="direct", aug=False, testlist_star_expr=True):
    """Add dependencies to <dependencies>
    <target> depends on <value>
    <target> may also depends on loop if <aug>
                                           or if there is a self_reference
    Expand assign dependencies if <testlist_star_expr>
    """

    left, right = AssignLeftVisitor(), AssignRightVisitor()

    if testlist_star_expr and tuple_or_list(target) and tuple_or_list(value):
        for i, targ in enumerate(target.elts):
            assign_dependencies(targ, value.elts[i], dependencies,
                                typ, dep_typ=dep_typ, testlist_star_expr=True)
        return

    left.visit(target)
    if value:
        right.visit(value)

    for name, _, lineno in left.names:
        var = variable(name, typ)
        dependencies[lineno][var]
        lineno = right.line if right.line != -1 else lineno
        self_reference = False
        for value, _, _ in right.names:
            dependencies[lineno][var].append(Dependency(value, dep_typ))
            if name == value:
                self_reference = True

        if aug:
            dependencies[lineno][var].append(Dependency(name, dep_typ))
            self_reference = True

        if self_reference:
            dependencies[lineno][var].append(Dependency("<self>", "loop"))


def assign_artificial_dependencies(target, artificial, dependencies, typ):
    """Add artificial dependencies to <dependencies>
    <target> depends on <artificial>
    """
    left = AssignLeftVisitor()

    left.visit(target)

    for name, _, _ in left.names:
        var = variable(name, typ)
        lineno = target.lineno
        dependencies[lineno][var].append(Dependency(artificial, "direct"))


def assign_right(lineno, target, value, dependencies):
    """Add dependencies to <dependencies>
    <target> depends on <value>
    """

    right = AssignRightVisitor()
    if value:
        right.visit(value)

    dependencies[lineno][target]
    for name, _, _ in right.names:
        dependencies[lineno][target].append(Dependency(name, "direct"))

class SlicingVisitor(FunctionVisitor):                                           # pylint: disable=too-many-instance-attributes, too-many-public-methods
    """Visitor that captures required information for program slicing"""

    def __init__(self, *args):
        super(SlicingVisitor, self).__init__(*args)
        self.line_usages = defaultdict(lambda: {
            "Load": [], "Store": [], "Del": [],
            "AugLoad": [], "AugStore": [], "Param": [],
        })
        self.dependencies = defaultdict(lambda: defaultdict(list))

        self.gen_dependencies = defaultdict(lambda: defaultdict(list))
        self.call_by_col = defaultdict(dict)
        self.function_calls_by_lasti = defaultdict(dict)
        self.with_enter_by_lasti = defaultdict(dict)
        self.with_exit_by_lasti = defaultdict(dict)
        self.function_calls_list = []
        self.with_list = []
        self.imports_list = []
        self.imports = set()
        self.iters_list = []
        self.iters = defaultdict(set)

        self.condition_stack = []
        self.conditions = {}
        self.loops = {}
        self.disasm = []

        # Python 2
        self.print_item_list = []
        self.print_newline_list = []

    def add_call_function(self, node, cls, *args, **kwargs):
        """Add special CallFunction of class <cls> to list
        Visit <node> to create dependencies
        Optional args: <call_list> is the list that will get the call
        """
        call_list = self.function_calls_list
        if "call_list" in kwargs and kwargs["call_list"] is not None:
            call_list = kwargs["call_list"]
        function_call = cls(AssignRightVisitor, *args)
        function_call.name = self.extract_code(node)
        function_call.visit(node)
        function_call.line, function_call.col = node.uid
        call_list.append(function_call)
        return function_call

    def add_decorators(self, node, original_name):
        """Add special function calls for decorators"""
        decorators = [self.add_decorator(dec_node)
                      for dec_node in node.decorator_list]
        lineno = node.lineno
        name = variable(original_name, 'call')
        for dec in reversed(decorators):
            uid = (dec.line, dec.col)
            self.dependencies[lineno][name].append(Dependency(uid, "direct"))
            #dec.args.append([name])
            name = variable(uid, 'call')

    def add_decorator(self, node):
        """Add special function call for decorator"""
        dec_node = ast.Call()
        dec_node.func = node
        dec_node.args = []
        dec_node.keywords = []
        if sys.version_info < (3, 5):
            dec_node.starargs = None
            dec_node.kwargs = None
        dec_node.first_col = node.first_col
        dec_node.first_line = node.first_line
        dec_node.last_col, dec_node.last_line = node.last_col, node.last_line
        dec_node.uid = (dec_node.first_line, dec_node.first_col)
        dec = self.add_call_function(dec_node, Decorator)
        self.call_by_col[dec.line][dec.col] = dec
        self.add_func_dependency(dec)
        return dec

    def add_new_comprehension(self, typ, node, add_call=True):
        """Create comprehension call and generators"""
        if add_call:
            call = self.add_call_function(node, GeneratorCall, typ)
            self.call_by_col[call.line][call.col] = call
            right = ComprehensionVisitor()
            right.visit(node)
            line, uid = call.line, variable((call.line, call.col), "return")
            for name, _, _ in right.names:
                self.dependencies[line][uid].append(Dependency(name, "direct"))
        for gen_node in node.generators:
            self.add_generator(typ, gen_node)

    def add_generator(self, typ, node):
        """Add special function call for generator"""
        self.add_call_function(node, Generator, typ, call_list=self.iters_list)

        assign_dependencies(node.target, node.iter, self.gen_dependencies,
                            "normal", testlist_star_expr=False)
        for nif in node.ifs:
            self.visit(nif)
            assign_dependencies(node.target, nif, self.gen_dependencies,
                                "conditional", testlist_star_expr=False)

    def add_with(self, node):
        """Cross version visit With and create dependencies"""
        _with = self.add_call_function(node, With, call_list=self.with_list)
        _with.line, _with.col = node.context_expr.uid
        if node.optional_vars:
            assign_artificial_dependencies(node.optional_vars,
                                           node.context_expr.uid,
                                           self.dependencies,
                                           "normal")
        self.call_by_col[_with.line][_with.col] = _with

    def add_return_yield(self, node, label):
        """Create special <label> variable
        Use for return and yield dependencies
        """
        self.new_var(ast.Name(label, ast.Store(), lineno=node.lineno),
                     "virtual", dep_typ="return", value=node.value)

    def new_var(self, name, typ, dep_typ="direct", value=None, **kwargs):
        """Create new variable"""
        assign_dependencies(name, value, self.dependencies, typ,
                            dep_typ=dep_typ,
                            testlist_star_expr=False, **kwargs)

    def add_func_dependency(self, call):
        """Create dependency for call nodes that accept expr as func"""
        for name in call.func:
            uid = variable((call.line, call.col), "call")
            self.dependencies[call.line][uid].append(
                Dependency(name, "direct"))

    def visit_stmts(self, stmts):
        """Visit stmts"""
        for stmt in stmts:
            self.visit(stmt)

    def visit_AugAssign(self, node):                                             # pylint: disable=invalid-name
        """Visit AugAssign. Create dependencies"""
        self.new_var(node.target, "normal", value=node.value, aug=True)
        self.generic_visit(node)

    def visit_Assign(self, node):                                                # pylint: disable=invalid-name
        """Visit Assign. Create dependencies"""
        for target in node.targets:
            assign_dependencies(target, node.value, self.dependencies,
                                "normal")
        self.generic_visit(node)

    def visit_For(self, node):                                                   # pylint: disable=invalid-name
        """Visit For. Create dependencies"""
        loop = Loop(node, "for")
        self.loops[loop.first_line] = loop
        _iter = self.add_call_function(node.iter, ForIter,
                                       call_list=self.iters_list)
        _iter.col += 1
        self.call_by_col[_iter.line][_iter.col] = _iter
        loop.maybe_call = (_iter.line, _iter.col)
        loop.add_iterable(node.iter, AssignRightVisitor)
        loop.add_iter_var(node.target, AssignLeftVisitor)
        loop.first_line_in_scope = node.body[0].first_line

        self.visit(node.target)
        self.visit(node.iter)
        self.visit_stmts(node.body)
        self.visit_stmts(node.orelse)

    def visit_AsyncFor(self, node):                                              # pylint: disable=invalid-name
        """Visit For. Create dependencies. Python 3.5"""
        self.visit_For(node)

    def visit_While(self, node):                                                 # pylint: disable=invalid-name
        """Visit While. Create conditional dependencies"""
        loop = Loop(node, "while")
        self.loops[loop.first_line] = loop
        self.visit_If(node)

    def visit_If(self, node):                                                    # pylint: disable=invalid-name
        """Visit If. Create conditional dependencies"""
        condition = Condition(node)
        self.conditions[condition.first_line] = condition
        condition.add_test(node.test, AssignRightVisitor)
        self.visit(node.test)
        self.condition_stack.append(condition)
        self.visit_stmts(node.body)
        self.visit_stmts(node.orelse)
        self.condition_stack.pop()

    def visit_Name(self, node):                                                  # pylint: disable=invalid-name
        """Visit Name. Crate Usage"""
        self.line_usages[node.lineno][type(node.ctx).__name__].append(node.id)
        super(SlicingVisitor, self).visit_Name(node)

    def visit_Call(self, node):                                                  # pylint: disable=invalid-name
        """Visit Call. Create special function call"""
        super(SlicingVisitor, self).visit_Call(node)
        call = self.add_call_function(node, FunctionCall)
        self.add_func_dependency(call)
        self.call_by_col[call.line][call.col] = call

    def visit_Print(self, node):                                                 # pylint: disable=invalid-name
        """Visit Print node. Create special function call
        Python 2 Only
        """
        self.generic_visit(node)
        _print = self.add_call_function(node, Print,
                                        call_list=self.print_newline_list)
        for _ in node.values:
            self.print_item_list.append(_print)
        self.call_by_col[_print.line][_print.col] = _print

    def visit_Return(self, node):                                                # pylint: disable=invalid-name
        """Visit Return. Create special variable"""
        for condition in self.condition_stack:
            condition.has_return = True
        self.add_return_yield(node, "return")
        if node.value:
            self.visit(node.value)

    def visit_Yield(self, node):                                                 # pylint: disable=invalid-name
        """Visit Yield. Create special variable"""
        self.add_return_yield(node, "yield")
        if node.value:
            self.visit(node.value)

    def visit_Import(self, node):                                                # pylint: disable=invalid-name
        """Visit Import"""
        self.generic_visit(node)

        for alias in node.names:
            _import = self.add_call_function(node, Import,
                                             call_list=self.imports_list)
            self.call_by_col[_import.line][_import.col] = _import

            name = ast.Name(alias.asname if alias.asname else alias.name,
                            ast.Store(), lineno=node.lineno)
            self.new_var(name, "import", value=node)

    def visit_ImportFrom(self, node):                                            # pylint: disable=invalid-name
        """Visit ImportFrom"""
        self.generic_visit(node)

        _import = self.add_call_function(node, Import,
                                         call_list=self.imports_list)
        self.call_by_col[_import.line][_import.col] = _import

        for alias in node.names:
            name = ast.Name(alias.asname if alias.asname else alias.name,
                            ast.Store(), lineno=node.lineno)
            self.new_var(name, "import from", value=node)

    def visit_ListComp(self, node):                                              # pylint: disable=invalid-name
        """Visit ListComp. Create special function call on Python 3"""
        self.add_new_comprehension("List", node, sys.version_info >= (3, 0))
        self.generic_visit(node)

    def visit_SetComp(self, node):                                               # pylint: disable=invalid-name
        """Visit SetComp. Create special function call"""
        self.add_new_comprehension("Set", node)
        self.generic_visit(node)

    def visit_DictComp(self, node):                                              # pylint: disable=invalid-name
        """Visit DictComp. Create special function call"""
        self.add_new_comprehension("Dict", node)
        self.generic_visit(node)

    def visit_GeneratorExp(self, node):                                          # pylint: disable=invalid-name
        """Visit GeneratorExp. Create special function call"""
        self.add_new_comprehension("Generator", node)
        self.generic_visit(node)

    def visit_ClassDef(self, node):                                              # pylint: disable=invalid-name
        """Visit ClassDef. Create special function call"""
        name = ast.Name(node.name, ast.Store(), lineno=node.lineno)
        name.first_line, name.first_col = name.uid = node.uid
        name.first_col += 1
        name.last_line = name.first_line
        name.last_col = name.first_col + len(node.name)
        _class = self.add_call_function(name, ClassDef)
        self.call_by_col[_class.line][_class.col] = _class

        line = node.lineno
        uid = (_class.line, _class.col)
        if not node.bases:
            assign_right(line, variable(uid, "call"), None, self.dependencies)


        for base in node.bases:
            assign_right(line, variable(uid, "call"), base, self.dependencies)


        self.add_decorators(node, uid)
        super(SlicingVisitor, self).visit_ClassDef(node)

    def visit_FunctionDef(self, node):                                           # pylint: disable=invalid-name
        """Visit FunctionDef"""
        name = ast.Name(node.name, ast.Store(), lineno=node.lineno)
        self.new_var(name, "function definition", value=None)

        self.add_decorators(node, node.name)
        super(SlicingVisitor, self).visit_FunctionDef(node)

    def visit_AsyncFunctionDef(self, node):                                      # pylint: disable=invalid-name
        """Visit AsyncFunctionDef. Python 3.5"""
        name = ast.Name(node.name, ast.Store(), lineno=node.lineno)
        self.new_var(name, "function definition", value=None)

        self.add_decorators(node, node.name)
        super(SlicingVisitor, self).visit_AsyncFunctionDef(node)

    def visit_Assert(self, node):                                                # pylint: disable=invalid-name
        """Visit Assert. Create special function call"""
        self.generic_visit(node)
        if node.msg:
            self.add_call_function(node, Assert, node.msg)

        # ToDo: with msg self.function_calls_list.append(cls)

    def visit_With(self, node):                                                  # pylint: disable=invalid-name
        """Visit With. Create special With enter/exit on Python 2"""
        self.generic_visit(node)
        if sys.version_info < (3, 0):
            self.add_with(node)

    def visit_withitem(self, node):
        """Visit Assert. Create special With enter/exit on Python 3"""
        self.generic_visit(node)
        self.add_with(node)

    def teardown(self):
        """Matches AST call order to call order in disassembly
        Possible issues:
        1- The disassembly may be specific to cpython. It may not work on other
        implementations
        2- If the order in AST is not correct, the matching will fail
        3- If there are other CALL_FUNCTION that are not considered in the AST
        the matching will fail
            both visit_ClassDef and visit_Call generates CALL_FUNCTION
        4- If some function is called without an explict CALL_FUNCTION
            __enter__: SETUP_WITH
            __exit__: WITH_CLEANUP
        """
        self.with_list.sort(key=lambda x: (x.line, x.col))
        end_with = {}
        operations = [
            AddCall(CALL_FUNCTIONS,
                    self.function_calls_list, self.function_calls_by_lasti),
            AddCall(PRINT_ITEMS,
                    self.print_item_list, self.function_calls_by_lasti),
            AddCall(PRINT_NEW_LINES,
                    self.print_newline_list, self.function_calls_by_lasti),
            AddImport(IMPORT_NAMES, self.imports,
                      self.imports_list, self.function_calls_by_lasti),
            AddWith(SETUP_WITH | SETUP_ASYNC_WITH, self.with_enter_by_lasti,
                    end_with, self.with_list, self.function_calls_by_lasti),
            AddWithCleanup(WITH_CLEANUP, self.with_exit_by_lasti,
                           end_with, self.function_calls_by_lasti),
            AddForIter(FOR_ITERS, self.iters,
                       self.iters_list, self.function_calls_by_lasti),
            AddGetIter(GET_ITERS, self.iters,
                       self.iters_list, self.function_calls_by_lasti),
        ]
        for inst in self.disasm:
            for operation in operations:
                operation.process(inst)


class AddCall(object):                                                           # pylint: disable=too-few-public-methods
    """Enrich disasm with calls and other instructions that cause calls"""

    def __init__(self, opcodes, clist, by_lasti):
        self.opcodes = opcodes
        self.clist = clist
        self.by_lasti = by_lasti
        self.index = 0

    def process(self, inst):
        """Process disasm instruction"""
        if inst.opcode in self.opcodes:
            call = safeget(self.clist, self.index)
            call.lasti = inst.offset
            self.by_lasti[inst.line][inst.offset] = call
            self.index += 1
            inst.extra = call
            return call


class AddImport(AddCall):                                                        # pylint: disable=too-few-public-methods
    """Enrich disasm with Imports"""

    def __init__(self, opcodes, imports, *args, **kwargs):
        super(AddImport, self).__init__(opcodes, *args, **kwargs)
        self.imports = imports

    def process(self, inst):
        """Process disasm instruction"""
        _import = super(AddImport, self).process(inst)
        if _import:
            self.imports.add(inst.line)


class AddWith(AddCall):                                                          # pylint: disable=too-few-public-methods
    """Enrich disasm with With statements"""

    def __init__(self, opcodes, with_enter_lasti, end_with, *args, **kwargs):
        super(AddWith, self).__init__(opcodes, *args, **kwargs)
        self.with_enter_by_lasti = with_enter_lasti
        self.end_with = end_with

    def process(self, inst):
        """Process disasm instruction"""
        _with = super(AddWith, self).process(inst)
        if _with:
            self.with_enter_by_lasti[inst.line][inst.offset] = _with
            end = int(inst.argrepr[3:])
            _with.end = end
            self.end_with[end] = _with
            return _with


class AddWithCleanup(namedtuple(
        "WithCleanup", "opcodes with_exit_by_lasti end_with by_lasti")):
    """Enrich disasm with With cleanups"""
    def process(self, inst):
        """Process disasm instruction"""
        if inst.opcode in self.opcodes:
            _with = self.end_with[inst.offset]
            del self.end_with[inst.offset]
            _with.end_line = inst.line
            self.by_lasti[inst.line][inst.offset] = _with
            self.with_exit_by_lasti[inst.line][inst.offset] = _with
            inst.extra = _with
            return _with


class AddForIter(AddCall):                                                        # pylint: disable=too-few-public-methods
    """Enrich disasm with ForIters"""

    def __init__(self, opcodes, iters, *args, **kwargs):
        super(AddForIter, self).__init__(opcodes, *args, **kwargs)
        self.iters = iters

    def process(self, inst):
        """Process disasm instruction"""
        _for = super(AddForIter, self).process(inst)
        if _for:
            self.iters[inst.line].add(inst.offset)


class AddGetIter(AddForIter):                                                     # pylint: disable=too-few-public-methods
    """Enrich disasm with GetIters"""

    def process(self, inst):
        """Process disasm instruction"""
        if sys.version_info < (3, 0):
            super(AddGetIter, self).process(inst)
        else:
            if inst.opcode in self.opcodes:
                self.iters[inst.line].add(inst.offset)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now dataflow'command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from ..persistence.models.graphs.dependency_graph import DependencyConfig
from ..persistence.models import Trial
from ..persistence import persistence_config

from .command import Command


class Dataflow(Command):
    """Export dataflow of a trial to dot file"""

    def add_arguments(self):
        add_arg = self.add_argument

        DependencyConfig.create_arguments(add_arg)
        add_arg("-v", "--value-length", type=int, default=0,
                help="R|maximum length of values (default: 0)\n"
                     "0 indicates that values should be hidden.\n"
                     "The values appear on the second line of node lables.\n"
                     "E.g. if it is set to '10', it will show 'data.dat',\n"
                     "but it will transform 'data2.dat' in to 'da...dat'\n"
                     "to respect the length restriction (note that '' is\n"
                     "part of the value).\n"
                     "Minimum displayable value: 5. Suggested: 55.")
        add_arg("-n", "--name-length", type=int, default=55,
                help="R|maximum length of names (default: 55)\n"
                     "0 indicates that values should be hidden.\n"
                     "Minimum displayable value: 5. Suggested: 55.")
        add_arg("trial", type=str, nargs="?",
                help="trial id or none for last trial")
        add_arg("--dir", type=str,
                help="set project path where is the database. Default to "
                     "current directory")

    def execute(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())
        trial = Trial(trial_ref=args.trial)
        trial.dependency_config.read_args(args)
        trial.dot.value_length = args.value_length
        trial.dot.name_length = args.name_length

        print(trial.dot.export_text())



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""History Graph Module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import time
import weakref

from collections import OrderedDict, defaultdict

from .structures import Graph
from ..trial import Trial


MAXTRIALS = 1000000
MAX_IN_GRAPH = float("inf")


class HistoryGraph(Graph):
    """History Graph Class

    Present history graph on Jupyter and on command line"""

    cache = {}

    def __init__(self, history, width=500, height=500):
        self.history = weakref.proxy(history)
        self.use_cache = True
        self.width = width
        self.height = height

    def history_data(self):
        """Create history data for presenting graphs

        Return:
        {"nodes": nodes, "edges": edges}
        nodes -- list of trials in filtered history
        edges -- list of edges dicts with keys source and target as node index
        """

        key = (self.history.script, self.history.status)
        if self.use_cache and key in self.cache:
            return self.cache[key]

        trials, ids, tmap = _preprocess_trials(Trial.reverse_trials(MAXTRIALS))
        graph, nodes, id_map, scripts = _prepare_history_graph(
            trials, ids, self.history.status.lower(), self.history.script
        )
        edges, order, children, actual_graph = _create_edges(
            graph, nodes, id_map
        )
        _set_trials_level(tmap, scripts, order, children, actual_graph)

        result = {"nodes": nodes, "edges": edges}
        if self.use_cache:
            self.cache[key] = result

        return result

    def graph(self):
        """Return history_data as a dict graph"""
        result = self.history_data()

        # To JSON
        final = []
        for trial in result["nodes"]:
            dic = trial.to_dict(
                ignore=tuple(),
                extra=("level", "status", "tooltip", "duration_text"))
            dic["start"] = str(dic["start"])
            dic["finish"] = str(dic["finish"])
            final.append(dic)
        return {"edges": result["edges"], "nodes": final}

    def _repr_html_(self):
        """Display d3 graph on ipython notebook"""
        uid = str(int(time.time() * 1000000))

        result = """
            <div class="nowip-history" data-width="{width}"
                 data-height="{height}" data-uid="{uid}">
                {data}
            </div>
        """.format(
            uid=uid,
            data=self.escape_json(self.graph()),
            width=self.width, height=self.height)
        return result

    def __repr__(self):
        width = max(len(s) for s in self.history.scripts)
        history = self.history_data()
        nodes = history["nodes"]

        if not nodes:
            return ""

        to_level = {
            nodes[edge["source"]].id: nodes[edge["target"]].level
            for edge in history["edges"]
        }

        max_level = max(t.level for t in nodes)
        active_levels = [0] * (max_level + 1)
        lines = []
        add_line = lambda *a, **kw: lines.append(_line_text(*a, **kw))

        for trial in reversed(nodes):
            active_levels[trial.level] = 1
            add_line(active_levels, trial, trial.level, width=width)

            if trial.id not in to_level:
                # First trial for script
                active_levels[trial.level] = 0

            elif to_level[trial.id] != trial.level:
                # Start of new branch
                active_levels[trial.level] = 0
                for i in range(trial.level, to_level[trial.id], -1):
                    add_line(active_levels, trial, i, moving=True, width=width)

        return "\n".join(lines)


def _preprocess_trials(trial_gen):
    """Preprocess trials


    Add level and tooltip to trials

    Return:
    trials -- preprocessed trials
    ids -- preprocessed trial ids
    tmap -- map trial.id to trial


    Arguments:
    trial_gen -- trial generator
    """
    trials = []
    ids = []
    tmap = {}

    for trial in trial_gen:
        trial.level = 0
        trial.tooltip = """
            <b>{0.script}</b><br>
            {status}<br>
            Start: {0.start}<br>
            Finish: {0.finish}
            """.format(trial, status=trial.status.capitalize())
        if trial.finish:
            trial.tooltip += """
            <br>
            Duration: {duration}
            """.format(duration=trial.duration_text)

        tmap[trial.id] = trial
        trials.append(trial)
        ids.append(trial.id)

    return (trials, ids, tmap)


def _prepare_history_graph(trials, ids, status, script):
    """Prepare history graph


    The graph is represented as a dict of dict of int
    Int values represent the distance between two nodes
    Use Floyd-Warshall algorithm to calculate distances


    This method also applies script and status filters on the graph
    Filters remove trials that do not match the conditions from the graph


    Return:
    graph -- distance graph
    nodes -- filtered trials
    id_map -- map trial.id to trial position in nodes list
    scripts -- group trials by scripts


    The returned graph can be used to:
    -find parent trial after filter
    -find previous trials in a branch line


    Arguments:
    trials -- trials list
    ids -- trial ids list
    """
    # Calculate distances
    graph = defaultdict(lambda: defaultdict(lambda: MAX_IN_GRAPH))

    for trial in trials:
        graph[trial.id][trial.id] = 0
        if trial.parent_id is not None:
            graph[trial.id][trial.parent_id] = 1

    for k in ids:
        for i in ids:
            for j in ids:
                if graph[i][j] > graph[i][k] + graph[k][j]:
                    graph[i][j] = graph[i][k] + graph[k][j]

    # Filter
    nodes = []
    id_map = {}
    scripts = defaultdict(list)
    nid = 0
    for trial in reversed(trials):
        if ((status != "*" and trial.status != status) or
                (script != "*" and trial.script != script)):
            for tid in ids:
                graph[tid][trial.id] = MAX_IN_GRAPH
        else:
            nodes.append(trial)
            scripts[trial.script].append(trial)
            id_map[trial.id] = nid
            nid += 1

    return (graph, nodes, id_map, scripts)


def _create_edges(graph, nodes, id_map):
    """Create edges for graph

    Arguments:
    graph -- dict of dict of int with min distances between trials
    nodes -- list of nodes from the oldest to the newest
    id_map -- map of trial.id to trial position in nodes list


    Return:
    edges -- edge list of dicts with source and target keys
    order -- ordered dict with desired script order
    children -- dict with list of trials that have key trial as target
    actual_graph -- edge dict by trial id
    """

    edges = []
    order = OrderedDict()
    children = defaultdict(list)
    actual_graph = {}

    for source, target in _edges(graph, nodes, script_order=order):
        edges.append({
            "source": id_map[source],
            "target": id_map[target],
            "right": 1,
            "level": 0
        })
        actual_graph[source] = target
        children[target].append(source)

    return (edges, order, children, actual_graph)


def _edges(graph, nodes, script_order=None):
    """Edge generator. Iterate through all edges on graph


    Arguments:
    graph -- dict of dict of int with min distances between trials
    nodes -- list of nodes from the oldest to the newest

    Keyword arguments:
    script_order -- ordered dict to be touched for setting the script order
    """
    if script_order is None:
        script_order = {}

    for trial in reversed(nodes):
        tid = trial.id
        target = min(
            graph[tid],
            key=lambda x, i=tid: float("inf") if x == i else graph[i][x]
        )
        if graph[tid][target] != MAX_IN_GRAPH:
            yield (tid, target)
        script_order[trial.script] = 1


def _set_trials_level(tmap, scripts, order, children, actual_graph):
    """Adjust levels of trials according to their script and branchs


    Arguments:
    tmap -- map of trial.id to trial
    scripts -- group trials by scripts
    order -- ordered dict with desired script order
    children -- dict with list of trials that have key trial as target
    actual_graph -- edge dict by trial id
    """

    level = 0
    for script in order:
        last = level
        for trial in scripts[script]:
            if trial.id not in actual_graph:
                # trial is isolated
                trial.level = level
                level += 1
                last += 1
                continue

            parent_id = actual_graph[trial.id]
            if children[parent_id].index(trial.id) > 0:
                # trial is not the first child
                # increase level
                trial.level = last
                last += 1
            else:
                # trial is the first child
                # use the parent level
                parent = tmap[parent_id]
                trial.level = parent.level
            level = max(level, trial.level + 1)


def _line_text(active, trial, current, moving=False, width=25):
    """Return text for line history"""
    text = []
    for i, value in enumerate(active):
        if i == current:
            text.append("/ " if moving else
                        " b" if not trial.run else
                        " *" if trial.finished else
                        " f")
        elif value:
            text.append(" |")
        else:
            text.append("  ")
    if moving:
        return "".join(text)
    return "{line}  {id: <4} {script: <{width}} {tags}".format(
        line="".join(text), id=trial.id, script=trial.script,
        tags=", ".join(tag.name for tag in trial.tags), width=width
    )



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define formatters for usual data structures to be used on IPython"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from collections import Counter


class PrettyStr(object):                                                         # pylint: disable=too-few-public-methods
    """Display lines with line breaks"""

    def _repr_pretty_(self, pretty, cycle):                                      # pylint: disable=unused-argument
        """Pretty repr for IPython"""
        pretty.text(str(self))


class PrettyLines(PrettyStr):                                                    # pylint: disable=too-few-public-methods
    """Display lines with line breaks"""

    def __init__(self, lines):
        """Constructor. lines must be a list of strings"""
        self.lines = lines

    def __str__(self):
        """Default str repr"""
        return "\n".join(self.lines)


class Table(list, PrettyStr):
    """Display tables"""

    def __init__(self, *args, **kwargs):
        super(Table, self).__init__(*args, **kwargs)
        self.show_header = True
        self.has_header = True

    def _repr_html_(self):
        """HTML repr for IPython"""
        result = "<table>"
        try:
            iterator = iter(self)
            if self.has_header:
                header = next(iterator)
                if self.show_header:
                    result += "<tr>"
                    result += "".join("<th>{}</th>".format(x) for x in header)
                    result += "</tr>"
            for row in iterator:
                result += "<tr>"
                result += "".join("<td>{}</td>".format(x) for x in row)
                result += "</tr>"
        except StopIteration:
            pass
        result += "</table><br>"
        return result

    def __str__(self):
        """Default str repr"""
        size = Counter()
        for row in self:
            for i, value in enumerate(row):
                size[i] = max(size[i], len(str(value)))
        result = ""
        try:
            iterator = iter(self)
            if self.has_header:
                header = next(iterator)
                if self.show_header:
                    result += " ".join("{0:>{1}}".format(x, size[i])
                                       for i, x in enumerate(header))
                    result += "\n"
            for row in iterator:
                result += " ".join("{0:>{1}}".format(x, size[i])
                                   for i, x in enumerate(row))
                result += "\n"
        except StopIteration:
            pass
        return result



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Debugger-related functions"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import atexit
import os
import subprocess
import sys
import threading
import webbrowser

from argparse import Namespace

from ...cmd.cmd_show import Show
from ...persistence.models.trial import Trial
from ...utils.io import redirect_output


FNULL = open(os.devnull, "w")
atexit.register(FNULL.close)

CHILDREN_PID = []


def kill_children():
    """Kill debugger processes"""
    global CHILDREN_PID                                                          # pylint: disable=global-statement
    for child in CHILDREN_PID:
        child.kill()
    CHILDREN_PID = []

atexit.register(kill_children)


def create_debugger(pdb=None):
    """Create debugger"""
    if pdb is None:
        with redirect_output():
            try:
                from ipdb.__main__ import Pdb, def_colors
                pdb = Pdb(def_colors)
            except ImportError:
                from pdb import Pdb
        pdb = Pdb

    class NowDebugger(pdb):
        """Debugger that invokes both pdb tracer and noWorkflow"""

        def __init__(self, provider, *args, **kwargs):
            super(NowDebugger, self).__init__(*args, **kwargs)
            self.provider = provider

        def trace_dispatch(self, frame, event, arg):
            pdb.trace_dispatch(self, frame, event, arg)
            self.provider.tracer(frame, event, arg)
            return self.trace_dispatch

    return NowDebugger


def debugger_builtins(provider, namespace, metascript):
    """Define debugging functions"""
    vis = vis_open, browser_open = [False, False]                                # pylint: disable=unused-variable

    def set_trace(frame=None, pdb=None):                                         # pylint: disable=unused-argument
        """Invoke pdb"""
        debugger = create_debugger(pdb)(provider)
        debugger.set_trace(sys._getframe().f_back)                               # pylint: disable=protected-access

    def now_save():
        """Save partial provenance"""
        provider.store(partial=True)

    def now_trial(trial_id=None, save=True):
        """Return current trial"""
        if save:
            now_save()
        if trial_id is None:
            trial_id = metascript.trial_id
        return Trial(trial_ref=trial_id)

    def now_vis(browser="maybe", port=5000, save=True, vis=vis):                 # pylint: disable=dangerous-default-value
        """Invoke now vis"""
        vis_open, browser_open = vis
        if save:
            now_save()
        url = "http://127.0.0.1:{0}".format(port)
        if vis_open:
            kill_children()
            vis[0] = vis_open = False
        if not vis_open:
            params = ["now", "vis", "-p", str(port)]
            print("now vis: {}".format(url))
            with redirect_output():
                vis[0] = True
                proc = subprocess.Popen(params, stdout=FNULL, stderr=FNULL)
                CHILDREN_PID.append(proc)

        if browser and (browser_open ^ (browser == "maybe")):
            def webopen(url):
                """Workaround to redirect browser output"""
                savout = os.dup(1)
                os.close(1)
                os.open(os.devnull, os.O_RDWR)
                try:
                    webbrowser.open(url)
                finally:
                    os.dup2(savout, 1)

            vis[1] = bool(browser)
            with redirect_output():
                threading.Timer(1.25, webopen, args=[url]).start()

    def now_notebook(start=True):
        """Start Jupyter Notebook"""
        try:
            import IPython
            Show().execute_export(Namespace(ipynb=True, dir=None))
            if start:
                params = ["ipython", "notebook", "Current Trial.ipynb"]
                proc = subprocess.Popen(params, stdout=FNULL, stderr=FNULL)
                CHILDREN_PID.append(proc)
        except ImportError:
            return "IPython not found"

    namespace["set_trace"] = set_trace
    namespace["now_save"] = now_save
    namespace["now_trial"] = now_trial
    namespace["now_vis"] = now_vis
    namespace["now_notebook"] = now_notebook

    try:
        with redirect_output():
            from IPython import embed
            namespace["now_ipython"] = embed
    except ImportError:
        namespace["now_ipython"] = lambda: "IPython not found"



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""AST Visitors to capture definition provenance"""

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import ast

from collections import defaultdict

from future.builtins import map as cvmap
from pyposast import extract_code

from ...utils.cross_version import cross_compile
from ...utils.bytecode.dis import instruction_dis_sorted_by_line


class FunctionVisitor(ast.NodeVisitor):                                          # pylint: disable=too-many-instance-attributes
    """Identifies the function declarations and related data"""

    def __init__(self, metascript, file_definition):
        self.path = file_definition.name
        self.code = file_definition.code
        self.lcode = self.code.split("\n")
        self.metascript = metascript
        self.result = None

        self.definitions = metascript.definitions_store
        self.objects = metascript.objects_store
        self.contexts = [file_definition]
        self.collecting_arguments = False
        self.function_globals = defaultdict(list)

        self.disasm = []

    def node_code(self, node):
        """Use PyPosAST positions to extract node text"""
        return extract_code(
            self.lcode, node, lstrip=" \t", ljoin="", strip="() \t"
        )

    def extract_code(self, node):
        """Use PyPosAST to extract node text without strips"""
        return extract_code(self.lcode, node)

    def new_definition_context(self, node, typ="FUNCTION"):
        """Visit node, open context. Collect code"""
        self.contexts.append(self.definitions.add_object(
            self.contexts[-1].namespace if len(self.contexts) > 1 else "",
            node.name,
            self.extract_code(node),
            typ,
            self.contexts[-1].id,
            node.first_line,
            node.last_line,
            ast.get_docstring(node)
        ))

        self.generic_visit(node)
        self.contexts.pop()

    def visit_ClassDef(self, node):                                              # pylint: disable=invalid-name
        """Visit ClassDef. Ignore Classes"""
        # ToDo #74: capture class dry_add -> add_object
        # ToDo #74: "".encode("utf-8") -> self.extract_code(node),
        # self.new_definition_context(node, typ="CLASS")
        self.contexts.append(self.definitions.dry_add(
            self.contexts[-1].namespace if len(self.contexts) > 1 else "",
            node.name,
            "",
            "CLASS",
            self.contexts[-1].id,
            node.first_line,
            node.last_line,
            ast.get_docstring(node)
        ))
        self.generic_visit(node)
        self.contexts.pop()

    def visit_FunctionDef(self, node):                                           # pylint: disable=invalid-name
        """Visit FunctionDef. Collect function code"""
        self.new_definition_context(node, typ="FUNCTION")

    def visit_AsyncFunctionDef(self, node):                                      # pylint: disable=invalid-name
        """Visit AsyncFunctionDef. Collect function code. Python 3.5"""
        self.new_definition_context(node, typ="FUNCTION")

    def visit_arguments(self, node):
        """Visit arguments. Collect arguments"""
        self.collecting_arguments = True
        self.generic_visit(node)
        self.collecting_arguments = False

    def visit_Global(self, node):                                                # pylint: disable=invalid-name
        """Visit Global. Collect globals"""
        definition = self.contexts[-1]
        for name in node.names:
            self.objects.add(name, "GLOBAL", definition.id)
            self.function_globals[definition.namespace].append(name)

        self.generic_visit(node)

    def call(self, node):
        """Collect direct function call"""
        func = node.func
        self.objects.add(
            self.node_code(func), "FUNCTION_CALL", self.contexts[-1].id)

    def visit_Call(self, node):                                                  # pylint: disable=invalid-name
        """Visit Call. Collect call"""
        self.call(node)
        self.generic_visit(node)

    def visit_Name(self, node):                                                  # pylint: disable=invalid-name
        """Visit Name. Get names"""
        if self.collecting_arguments:
            self.objects.add(node.id, "ARGUMENT", self.contexts[-1].id)
        self.generic_visit(node)

    def teardown(self):
        """Disable"""
        pass

    def extract_disasm(self):
        """Extract disassembly code"""
        compiled = cross_compile(
            self.code, self.path, "exec")
        if self.path == self.metascript.path:
            self.metascript.compiled = compiled

        self.disasm = instruction_dis_sorted_by_line(compiled, recurse=True)
        if self.metascript.disasm0:
            print("------------------------------------------------------")
            print(self.path)
            print("------------------------------------------------------")
            print("\n".join(cvmap(repr, self.disasm)))
            print("------------------------------------------------------")



import csv
import sys
import matplotlib.pyplot as plt
from simulator import simulate

def run_simulation(data_a, data_b):
    a = csv_read(data_a)
    b = csv_read(data_b)
    data = simulate(a, b)
    return data

def csv_read(f):
    reader = csv.reader(open(f, 'rU'), delimiter=':')
    data = []
    for row in reader:
        data.append(row)
    return data

def extract_column(data, column):
    col_data = []
    for row in data:
        col_data.append(float(row[column]))
    return col_data

def extract_temperature(data):
    return extract_column(data, 0)

def extract_precipitation(data):
    return extract_column(dat, 1)

def plot(data):
    t = extract_temperature(data)
    p = extract_precipitation(data)
    plt.scatter(t, p, marker='o')
    plt.xlabel('Temperature')
    plt.ylabel('Precipitation')
    plt.savefig("output.png")

#Main Program
data_a = sys.argv[1]
data_b = sys.argv[2]
data = run_simulation(data_a, data_b)
plot(data)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Helpers for Prolog Extraction"""


from datetime import datetime


class PrologDescription(object):
    """Prolog Description. Generate comments, facts, dynamic, and retract"""

    def __init__(self, name, attributes, description=""):
        self.name = name
        self.attributes = attributes
        self.description = description.split("\n")

    def comment(self):
        """Return prolog comment"""
        result = []
        result.append("")
        result.append("%")
        result.append("% FACT DEFINITION: {}".format(repr(self)))
        if len(self.description) > 1 or self.description[0]:
            result.append("% DESCRIPTION: {}".format(self.description[0]))
            for desc in self.description[1:]:
                result.append("%              {}".format(desc))
        result.append("%")
        result.append("")
        return "\n".join(result)

    def dynamic(self):
        """Return prolog dynamic clause"""
        return ":- dynamic({0.name}/{1}).".format(self, len(self.attributes))

    def retract(self, trial_id):
        """Return prolog retract for trial"""
        return "retract({0.name}({1}))".format(
            self, ', '.join(x.retract(trial_id) for x in self.attributes)
        )

    def fact(self, obj):
        """Convert obj to prolog fact"""
        return "{0.name}({1}).".format(
            self, ', '.join(x.fact(obj) for x in self.attributes)
        )

    def empty(self):
        """Return empty fact"""
        return "{0.name}({1}).".format(
            self, ', '.join(x.empty() for x in self.attributes)
        )

    def __repr__(self):
        return "{0.name}({1}).".format(
            self, ', '.join(x.variable() for x in self.attributes)
        )


class PrologAttribute(object):
    """Represent a single attribute"""

    def __init__(self, name, fn=None, attr_name=None, link=None):
        self.name = name
        self.attr_name = self.name if attr_name is None else attr_name
        self.func = fn
        self.link = link

    def variable(self):
        return "".join(x.title() for x in self.name.split("_"))

    def value(self, obj):
        """Return attribute self.attr_name of obj"""
        if self.func:
            return self.func(obj)
        attr = self.attr_name
        while "." in attr:
            attr0, attr = attr.split(".", 1)
            obj = getattr(obj, attr0)
        return getattr(obj, attr)

    def retract(self, trial_id):                                                 # pylint: disable=unused-argument, no-self-use
        """Attribute does not identify fact. Retrun _"""
        return "_"

    def fact(self, obj):
        """Return attribute self.attr_name of obj as str"""
        return str(self.value(obj))

    def empty(self):                                                             # pylint: disable=no-self-use
        """Represent empty attribute"""
        return "0"


class PrologTrial(PrologAttribute):
    """Represent Trial id attribute, used for retract"""

    def retract(self, trial_id):
        """Trial Attribute identifies fact. Retrun trial_id"""
        return str(trial_id)


class PrologRepr(PrologAttribute):
    """Represent an attribute that should be written with quotes"""

    def fact(self, obj):
        """Return attribute self.attr_name of obj as escaped repr"""
        result = repr(self.value(obj))
        if result[0] not in ('"', "'") and result[1] in ('"', "'"):
            result = result[1:]
        if result[0] in ('"', "'") and result[-1] == result[0]:
            result = result[1:-1]
        result = result.replace("'", "''")
        return "'{}'".format(result)


class PrologTimestamp(PrologAttribute):
    """Represent a timestamp"""

    def fact(self, obj):
        """Return attribute self.attr_name of obj as formatted timestamp"""
        time = self.value(obj)
        if not time:
            return -1
        epoch = datetime(1970, 1, 1)
        return str((time - epoch).total_seconds())


class PrologNullable(PrologAttribute):
    """Represent an attribute that accepts nil as value"""

    def fact(self, obj):
        """Replace None by nil if attribute self.attr_name of obj"""
        value = self.value(obj)
        return str(value) if value else "nil"


class PrologNullableRepr(PrologRepr):
    """Represent an attribute that accepts nil as value and requires quotes"""

    def fact(self, obj):
        """Return attribute self.attr_name of obj as escaped repr"""
        if not self.value(obj):
            return "nil"
        return super(PrologNullableRepr, self).fact(obj)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Diagrams module. Define both Schema and Prolog Diagram"""


def viz_table(body, border=0, cellborder=1, cellpadding=2, cellspacing=0,        # pylint: disable=too-many-arguments, unused-argument
              bgcolor="white", color="dodgerblue3"):                             # pylint: disable=unused-argument
    """Create HTML table for graphviz"""
    return (
        '<TABLE BORDER="{border}" CELLBORDER="{cellborder}"'
        ' CELLSPACING="{cellspacing}" CELLPADDING="{cellpadding}"'
        ' BGCOLOR="{bgcolor}" COLOR="{color}">'
        '{body}'
        '</TABLE>'
    ).format(**locals())


def viz_tr(body, port=None):
    """Create HTML tr for graphviz"""
    if port:
        return '<TR PORT="{}">{}</TR>'.format(port, body)
    return '<TR>{}</TR>'.format(body)


def viz_td(body, port=None, align="CENTER"):                                     # pylint: disable=unused-argument
    """Create HTML td for graphviz"""
    port_text = "" if port is None else 'PORT="{}"'.format(port)
    return '<TD {port_text} ALIGN="{align}">{body}</TD>'.format(**locals())


def viz_b(body):
    """Create HTML b for graphviz"""
    return '<B>{}</B>'.format(body)


def viz_white(body, bgcolor="white"):                                            # pylint: disable=unused-argument
    """Create hidden text for graphviz"""
    return '<FONT COLOR="{bgcolor}">{body}</FONT>'.format(**locals())


def viz_white_wrap(body, char, bgcolor="white"):                                 # pylint: disable=unused-argument
    """Wrap body with hidden text for graphviz"""
    return (
        '<FONT COLOR="{bgcolor}">'
        ' <FONT COLOR="black">{body}</FONT>'
        '{char}</FONT>'
    ).format(**locals())


def viz_property(statement, properties):
    """Create properties for graphviz element"""
    if not properties:
        return statement + ";";
    return statement + "[{}];".format(" ".join(properties))


class ViewDiagram(object):
    """ViewDiagram Base Class"""

    def __init__(self, _format="svg"):
        self.points = 0
        self.format = _format
        self.output = ""
        self.links = []
        self.result = []

    def _repr_svg_(self):
        if self.format == "svg":
            ipython = get_ipython()
            return ipython.run_cell_magic(
                "dot", "--format {}".format(self.format), self.as_dot()
            )

    def _repr_png_(self):
        if self.format == "png":
            ipython = get_ipython()
            return ipython.run_cell_magic(
                "dot", "--format {}".format(self.format), self.as_dot()
            )

    def as_dot(self):                                                            # pylint: disable=no-self-use
        """Export diagram to dot"""
        ...

    def __enter__(self):
        self.links = []
        self.points = 0
        self.result = ["digraph G{"]
        self.result.append("    rankdir=LR;")
        self.result.append("    overlap=false;")
        self.result.append("    splines=polyline;")
        return self.result

    def __exit__(self, *args):
        self.result += self.links
        self.result.append("}")
        self.output = "\n".join(self.result)


class ViewPrologDiagram(ViewDiagram):
    """View Prolog Diagram"""

    def __init__(self, descriptions, _format="svg"):
        super(ViewPrologDiagram, self).__init__(_format=_format)
        self.description = descriptions

    def add_link(self, model, link, attr_name):
        """Create edge"""
        if link:
            splitted = link.split('.')
            source = "{}:{}".format(model, attr_name)
            target = "{}:{}".format(*splitted)

            if splitted[0] == model:
                link_format = "    {0}:e -> p_{1} -> {2}:e".format(
                    source, self.points, target
                )
                point = viz_property(
                    "    p_{}".format(self.points),
                    ["shape=point", "color=dodgerblue3"]
                )
                self.links.append(point)
                self.points += 1
            else:
                link_format = "    {}:_ -> {}:_".format(source, target)

            self.links.append(viz_property(
                link_format,
                ["dir=none", "color=dodgerblue3"]
            ))

    def as_dot(self):
        with self as result:
            for model in self.description:
                content = [
                    viz_tr(viz_td(viz_white_wrap(viz_b(model.name), "(")))
                ]
                for attr in model.attributes[:-1]:
                    content.append(viz_tr(
                        viz_td(viz_white_wrap(attr.variable(), ","),
                               port=attr.name)
                    ))
                    self.add_link(model.name, attr.link, attr.name)
                attr = model.attributes[-1]
                content.append(viz_tr(
                    viz_td(viz_white_wrap(attr.variable(), ")"),
                           port=attr.name)
                ))
                self.add_link(model.name, attr.link, attr.name)
                result.append('    {0} [shape=none label=<{1}>];'.format(
                    model.name, viz_table("".join(content))
                ))
        return self.output


class ViewRelationalDiagram(ViewDiagram):
    """View SQL Diagram"""

    def __init__(self, order, _format="svg"):
        super(ViewRelationalDiagram, self).__init__(_format=_format)
        self.order = order

    def add_link(self, model, link, attr_name):
        """Create arrow"""
        link_properties = ['label="{}"'.format(attr_name), "color=dodgerblue3"]
        splitted = link.split('.')
        source = model
        target = "{}:{}".format(*splitted)

        if splitted[0] == model:
            link_format = "    {0}:n -> {1}:e".format(source, target)
        else:
            link_format = "    {0}:_ -> {1}:_".format(source, target)

        self.links.append(viz_property(link_format, link_properties))

    def as_dot(self):
        with self as result:
            for model in self.order:
                name = model.__tablename__
                table = model.t
                content = [viz_tr(viz_td(viz_b(name)))]
                attributes = []
                for column in table.columns:
                    if not column.foreign_keys:
                        attributes.append(viz_tr(viz_td(
                            "{}: {!s}".format(column.name, column.type),
                            align="LEFT", port=column.name
                        )))
                    else:
                        for fkey in column.foreign_keys:
                            if not fkey.column.foreign_keys:
                                link = fkey.target_fullname
                                self.add_link(name, link, column.name)
                if attributes:
                    content.append(viz_tr(viz_td(viz_table(
                        "".join(attributes), border=0, cellborder=0
                    ))))
                else:
                    content.append(viz_tr(viz_td("")))
                result.append('    {0} [shape=none label=<{1}>];'.format(
                    name,
                    viz_table("".join(content), cellspacing=0)
                ))
        return self.output



import sys
from external import align_warp, reslice, softmean, slicer
from convert import pgmtoppm, pnmtojpeg

def align_reslice(anatomy_image, reference_image):
    anatomy_header = anatomy_image[:-3] + "hdr"
    reference_header = reference_image[:-3] + "hdr"
    warp = align_warp(anatomy_image, anatomy_header,
                      reference_image, reference_header)
    return reslice(warp)

def slice_convert(atlas_image, atlas_header, coordinate):
    atlas_slice = slicer(atlas_image, atlas_header, coordinate)
    return pnmtojpeg(pgmtoppm(atlas_slice))

def main():
    reference = sys.argv[-1]
    anatomy_images = sys.argv[1:-1]
    resliced = []
    for anatomy in anatomy_images:
        resliced += align_reslice(anatomy, reference)
    atlas_image, atlas_header = softmean(*resliced)
    for coordinate in ["x", "y", "z"]:
        atlas = slice_convert(atlas_image, atlas_header, coordinate)

main()



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""General persistence Configurations"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys

from os.path import join, isdir

from ..utils.io import print_msg


PROVENANCE_DIRNAME = ".noworkflow"


class PersistenceConfig(object):
    """Config persistence. Interface for connecting to directory"""

    def __init__(self, path=None, connect=False):
        self.base_path = None  # Exeution path
        self.provenance_path = None  # Base .noworkflow path
        self.db_conn = None  # Connection to the database
        self.should_mock = False

        if path:
            self.path = path

        if connect:
            self.connect(path)

        self.delegate = []

    @property
    def path(self):
        """Return database path"""
        return self.base_path

    @path.setter
    def path(self, path):
        self.base_path = path
        self.provenance_path = join(path, PROVENANCE_DIRNAME)

        for obj in self.delegate:
            obj.set_path(self)

    def _has_provenance(self, path=None):
        """Check if persistence path exists"""
        if path:
            return isdir(join(path, PROVENANCE_DIRNAME))
        return isdir(self.provenance_path)

    def mock(self):
        """Mock database access"""
        self.should_mock = True
        for obj in self.delegate:
            obj.mock(self)

    def connect(self, path=None):
        """Connect to persistence"""
        if path:
            self.path = path

        for obj in self.delegate:
            obj.connect(self)

    def connect_existing(self, path=None):
        """Connect to existing persistence. Exit if it does not exist."""
        if path:
            self.path = path
        if not self._has_provenance():
            print_msg("there is no provenance store in the current directory",
                      True)
            sys.exit(1)
        self.connect()

    def add(self, obj):
        """Add database manager to config for delegation"""
        self.delegate.append(obj)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Definition provenance collector. Handle multiple files/visitors"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


import weakref

from collections import defaultdict

from future.builtins import map as cvmap
from future.utils import viewitems

import pyposast

from .slicing_visitor import SlicingVisitor

from ...persistence.models import FunctionDef, Object
from ...utils.io import print_msg
from ...utils.metaprofiler import meta_profiler


class Definition(object):                                                        # pylint: disable=too-many-instance-attributes
    """Collect definition provenance"""

    def __init__(self, metascript):
        self.metascript = weakref.proxy(metascript)
        self.paths = []
        # Map of dependencies by line
        self.line_dependencies = {}
        # Map of dependencies by line
        self.line_gen_dependencies = {}
        # Map of loops by line
        self.loops = {}
        # Map of conditional statements (if, while) by line
        self.conditions = {}
        # Map of name_refs by line
        self.line_usages = {}
        # Map of calls by line and col
        self.call_by_col = {}
        # Map of calls by offset line and lasti
        self.call_by_lasti = {}
        # Map of with __enter__ by line and lasti
        self.with_enter_by_lasti = {}
        # Map of with __exit__ by line and lasti
        self.with_exit_by_lasti = {}
        # Set of imports
        self.imports = {}
        # Set of GET_ITER and FOR_ITER lasti by line
        self.iters = {}
        # Function definitions
        self.function_globals = defaultdict(lambda: defaultdict(list))

    @meta_profiler("definition")
    def collect_provenance(self):
        """Collect definition provenance from scripts in metascript.paths"""
        metascript = self.metascript
        print_msg("  registering user-defined functions")
        for path, file_definition in viewitems(metascript.paths):
            visitor = self._visit_ast(file_definition)
            if visitor:
                if metascript.disasm:
                    print("--------------------------------------------------")
                    print(path)
                    print("--------------------------------------------------")
                    print("\n".join(cvmap(repr, visitor.disasm)))
                    print("--------------------------------------------------")
                self._add_visitor(visitor)

    def store_provenance(self):
        """Store definition provenance"""
        metascript = self.metascript
        tid = metascript.trial_id
        # Remove after save
        partial = True
        FunctionDef.fast_store(tid, metascript.definitions_store, partial)
        Object.fast_store(tid, metascript.objects_store, partial)

    def _visit_ast(self, file_definition):
        """Return a visitor that visited the tree"""
        metascript = self.metascript
        try:
            tree = pyposast.parse(file_definition.code, file_definition.name)
        except SyntaxError:
            print_msg("Syntax error on file {}. Skipping file.".format(
                file_definition.name))
            return None

        visitor = SlicingVisitor(metascript, file_definition)
        visitor.result = visitor.visit(tree)
        visitor.extract_disasm()
        visitor.teardown()
        return visitor

    def _add_visitor(self, visitor):
        """Add visitor data to Definition object"""
        self.paths.append(visitor.path)
        self.line_dependencies[visitor.path] = visitor.dependencies
        self.line_gen_dependencies[visitor.path] = visitor.gen_dependencies
        self.line_usages[visitor.path] = visitor.line_usages
        self.call_by_col[visitor.path] = visitor.call_by_col
        self.call_by_lasti[visitor.path] = visitor.function_calls_by_lasti
        self.with_enter_by_lasti[visitor.path] = visitor.with_enter_by_lasti
        self.with_exit_by_lasti[visitor.path] = visitor.with_exit_by_lasti
        self.imports[visitor.path] = visitor.imports
        self.iters[visitor.path] = visitor.iters
        self.function_globals[visitor.path] = visitor.function_globals
        self.loops[visitor.path] = visitor.loops
        self.conditions[visitor.path] = visitor.conditions



import csv
import time
import tempfile
import os
#this code requires matplotlib to be installed
#please refer to http://matplotlib.org/downloads.html
import matplotlib.pyplot as plt

class TabularData:
    def __init__(self, data, header=None):
        self.data = data
        self.header = header

###############################################################

def simulation(dataA, dataB):
    time.sleep(2)

    all_data = dataA.data + dataB.data

    (f, name) = tempfile.mkstemp(prefix='vtweather')
    os.close(f)

    writer = csv.writer(open(name, 'w'), delimiter=':')
    writer.writerows(all_data)

    #res = registry.get_descriptor_by_name('edu.utah.sci.vistrails.basic',
    #    'File').module()
    #res.name = name
    return name


def csvRead(f):
    reader = csv.reader(open(f, 'rU'), delimiter=':')
    header = []
    data = []
    for row in reader:
        data.append(row)
    tab_data = TabularData(data, header)
    #self.setResult('data', tab_data)
    return tab_data

def extractColumn(d, column_name, column):
        data = d.data
        header = d.header

        #Test to check whether column_name is not empty
        if column_name:
            if header is None:
                print("Data does not contain header")
            #column_name = self.getInputFromPort('columnName')
            try:
                idx = header.index(column_name)
            except ValueError:
                print("Data does not contain column" + column)
        else:
                idx = column

        col_data = []
        for row in data:
            print('processing row', row)
            col_data.append(row[idx])
        return col_data

def mplScatter(x,y):
    kwargs = {}
    #kwargs['y'] = y
    #kwargs['x'] = x
    plt.scatter(x, y, s=20, c='b', marker='o', cmap=None, norm=None,
                    vmin=None, vmax=None, alpha=None, linewidths=None, verts=None, **kwargs)
    plt.xlabel('Temperature')
    plt.ylabel('Precipitation')
    plt.savefig("output.png")

######################################################################################
#Main Program
print('Reading data data1.dat...')
dataA = csvRead('data1.dat')
print('Reading data data2.dat...')
dataB = csvRead('data2.dat')
#Simulation
print('Executing simulation...')
tempFile = simulation(dataA, dataB)
#tempFile was generated by Simulation
print('Reading temporary file...')
d = csvRead(tempFile)
#GetPrecipitation
print('Extracting precipitation...')
columnY = extractColumn(d, '', 1)
#GetTemperature
print('Extracting temperature...')
columnX = extractColumn(d, '', 0)
#Transform column values in a list of float
print('Converting values...')
out1 = [float(i) for i in columnX]
out2 = [float(i) for i in columnY]
print('Generating result...')
mplScatter(out1, out2)





# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Trial Prolog Object"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import weakref

from ...utils.functions import resource

from .base import Model
from .graphs.diagram import ViewPrologDiagram
from . import Tag, Dependency, EnvironmentAttr
from . import FunctionDef, Object
from . import Activation, FileAccess, ObjectValue
from . import Variable, VariableUsage, VariableDependency


RULES = "../resources/rules.pl"


class TrialProlog(Model):
    """Handle Prolog export and SWIPL integration"""

    __modelname__ = "TrialProlog"
    prolog_cli = None

    def __init__(self, trial):
        super(TrialProlog, self).__init__()
        self.use_cache = True
        self.trial = weakref.proxy(trial)

        self.models = self.prolog_models(trial)

    @classmethod
    def prolog_models(cls, trial=None):
        """Prolog models and accessors"""
        from . import Trial
        return [
            (Trial, lambda: [trial]),
            (Tag, lambda: trial.tags),
            (Dependency, lambda: trial.dependencies),
            (EnvironmentAttr, lambda: trial.environment_attrs),
            (FunctionDef, lambda: trial.function_defs),
            (Object, lambda: trial.objects),
            (Activation, lambda: trial.activations),
            (ObjectValue, lambda: trial.object_values),
            (FileAccess, lambda: trial.file_accesses),
            (Variable, lambda: trial.prolog_variables.variables),
            (VariableUsage, lambda: trial.prolog_variables.usages),
            (VariableDependency, lambda: trial.prolog_variables.dependencies),
        ]

    @classmethod
    def diagram(cls, format_="svg"):
        """Show prolog diagram"""
        descriptions = [x.prolog_description for x, _ in cls.prolog_models()]
        return ViewPrologDiagram(descriptions, format_)

    def retract(self):
        """Remove extracted rules from swipl"""
        for cls, _ in self.models:
            list(self.prolog_cli.query(
                cls.prolog_description.retract(self.trial.id)))

    def _export_facts(self, with_doc=True):
        """Export facts from trial as a list"""
        result = []
        for cls, query in self.models:
            if with_doc:
                result.append(cls.prolog_description.comment())
            result.append(cls.prolog_description.dynamic())
            for obj in query():
                result.append(cls.prolog_description.fact(obj))
        return result

    def export_text_facts(self):
        """Export facts from trial as text"""
        return "\n".join(self._export_facts())

    def rules(self, with_facts=False):
        """Export prolog rules


        Keywork arguments:
        with_facts -- export fact description too (default = False)"""
        result = []
        if with_facts:
            for cls, _ in self.models:
                result.append(cls.prolog_description.comment())
        result += resource(RULES, "UTF-8").split("\n")
        return result

    def load_cli_facts(self):
        """Load prolog facts an rules into swipl"""
        self.init_cli()
        load_trial = self.trial.prolog_description.fact(self.trial)[:-1]
        if not list(self.prolog_cli.query(load_trial)):
            for fact in self._export_facts(with_doc=False):
                self.prolog_cli.assertz(fact[:-1])
        load_rules = "load_rules(1)"
        if not list(self.prolog_cli.query(load_rules)):
            rule = []
            for line in self.rules():
                line = line.strip()
                if not line or line[0] == "%":
                    continue
                rule.append(line)
                if "." in line:
                    self.prolog_cli.assertz(" ".join(rule)[:-1])
                    rule = []
            self.prolog_cli.assertz(load_rules)

    def query(self, query):
        """Run prolog query on trial"""
        self.load_cli_facts()
        return self.prolog_query(query)

    def __hash__(self):
        return self.trial.id

    @classmethod
    def init_cli(cls):
        """Initialize swipl if it was not initialized yet"""
        # Avoid import loop
        from . import Trial

        if not cls.prolog_cli:
            from pyswip import Prolog
            cls.prolog_cli = Prolog()
            cls.prolog_cli.assertz(Trial.prolog_description.empty()[:-1])
            cls.prolog_cli.assertz("load_rules(0)")

    @classmethod
    def prolog_query(cls, query):
        """Run prolog query without specifying trial"""
        cls.init_cli()

        id_to_instance = {}
        cache = set()
        no_cache = set()

        for inst in cls.get_instances():                                         # pylint: disable=no-member
            id_to_instance[inst.trial.id] = inst
            (cache if inst.use_cache else no_cache).add(inst.trial.id)
        retract_ids = no_cache - cache
        for tid in retract_ids:
            id_to_instance[tid].retract()

        return cls.prolog_cli.query(query)
